# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.locations
~~~~~~~~~~~~~~
Provides location lookup dictionaries

Attributes:
    LOCATIONS (dict): Location name to info mapping
"""
LOCATIONS = {
    "Asia": {"code_2": "AS", "continent": "Asia"},
    "Africa": {"code_2": "AF", "continent": "Africa"},
    "Europe": {"code_2": "EU", "continent": "Europe"},
    "North America": {"code_2": "NA", "continent": "North America"},
    "South America": {"code_2": "SA", "continent": "South America"},
    "Oceania": {"code_2": "OC", "continent": "Oceania"},
    "Afghanistan": {
        "code_2": "AF",
        "code_3": "AFG",
        "continent": "Asia",
        "country": "Afghanistan",
        "num": "4",
    },
    "Albania": {
        "code_2": "AL",
        "code_3": "ALB",
        "continent": "Europe",
        "country": "Albania",
        "num": "8",
    },
    "Algeria": {
        "code_2": "DZ",
        "code_3": "DZA",
        "continent": "Africa",
        "country": "Algeria",
        "num": "12",
    },
    "American Samoa": {
        "code_2": "AS",
        "code_3": "ASM",
        "continent": "Oceania",
        "country": "American Samoa",
        "num": "16",
    },
    "Andorra": {
        "code_2": "AD",
        "code_3": "AND",
        "continent": "Europe",
        "country": "Andorra",
        "num": "20",
    },
    "Angola": {
        "code_2": "AO",
        "code_3": "AGO",
        "continent": "Africa",
        "country": "Angola",
        "num": "24",
    },
    "Anguilla": {
        "code_2": "AI",
        "code_3": "AIA",
        "continent": "North America",
        "country": "Anguilla",
        "num": "660",
    },
    "Antarctica": {
        "code_2": "AQ",
        "code_3": "ATA",
        "continent": "Antarctica",
        "country": "Antarctica",
        "num": "10",
    },
    "Antigua and Barbuda": {
        "code_2": "AG",
        "code_3": "ATG",
        "continent": "North America",
        "country": "Antigua and Barbuda",
        "num": "28",
    },
    "Argentina": {
        "code_2": "AR",
        "code_3": "ARG",
        "continent": "South America",
        "country": "Argentina",
        "num": "32",
    },
    "Armenia": {
        "code_2": "AM",
        "code_3": "ARM",
        "continent": "Europe",
        "country": "Armenia",
        "num": "51",
    },
    "Aruba": {
        "code_2": "AW",
        "code_3": "ABW",
        "continent": "North America",
        "country": "Aruba",
        "num": "533",
    },
    "Australia": {
        "code_2": "AU",
        "code_3": "AUS",
        "continent": "Oceania",
        "country": "Australia",
        "num": "36",
    },
    "Austria": {
        "code_2": "AT",
        "code_3": "AUT",
        "continent": "Europe",
        "country": "Austria",
        "num": "40",
    },
    "Azerbaijan": {
        "code_2": "AZ",
        "code_3": "AZE",
        "continent": "Europe",
        "country": "Azerbaijan",
        "num": "31",
    },
    "Bahamas": {
        "code_2": "BS",
        "code_3": "BHS",
        "continent": "North America",
        "country": "Bahamas",
        "num": "44",
    },
    "Bahrain": {
        "code_2": "BH",
        "code_3": "BHR",
        "continent": "Asia",
        "country": "Bahrain",
        "num": "48",
    },
    "Bangladesh": {
        "code_2": "BD",
        "code_3": "BGD",
        "continent": "Asia",
        "country": "Bangladesh",
        "num": "50",
    },
    "Barbados": {
        "code_2": "BB",
        "code_3": "BRB",
        "continent": "North America",
        "country": "Barbados",
        "num": "52",
    },
    "Belarus": {
        "code_2": "BY",
        "code_3": "BLR",
        "continent": "Europe",
        "country": "Belarus",
        "num": "112",
    },
    "Belgium": {
        "code_2": "BE",
        "code_3": "BEL",
        "continent": "Europe",
        "country": "Belgium",
        "num": "56",
    },
    "Belize": {
        "code_2": "BZ",
        "code_3": "BLZ",
        "continent": "North America",
        "country": "Belize",
        "num": "84",
    },
    "Benin": {
        "code_2": "BJ",
        "code_3": "BEN",
        "continent": "Africa",
        "country": "Benin",
        "num": "204",
    },
    "Bermuda": {
        "code_2": "BM",
        "code_3": "BMU",
        "continent": "North America",
        "country": "Bermuda",
        "num": "60",
    },
    "Bhutan": {
        "code_2": "BT",
        "code_3": "BTN",
        "continent": "Asia",
        "country": "Bhutan",
        "num": "64",
    },
    "Bolivia": {
        "code_2": "BO",
        "code_3": "BOL",
        "continent": "South America",
        "country": "Bolivia",
        "num": "68",
    },
    "Bonaire": {
        "code_2": "BQ",
        "code_3": "BES",
        "continent": "North America",
        "country": "Bonaire",
        "num": "535",
    },
    "Bosnia and Herzegovina": {
        "code_2": "BA",
        "code_3": "BIH",
        "continent": "Europe",
        "country": "Bosnia and Herzegovina",
        "num": "70",
    },
    "Botswana": {
        "code_2": "BW",
        "code_3": "BWA",
        "continent": "Africa",
        "country": "Botswana",
        "num": "72",
    },
    "Bouvet Island": {
        "code_2": "BV",
        "code_3": "BVT",
        "continent": "Antarctica",
        "country": "Bouvet Island",
        "num": "74",
    },
    "Brazil": {
        "code_2": "BR",
        "code_3": "BRA",
        "continent": "South America",
        "country": "Brazil",
        "num": "76",
    },
    "British Virgin Islands": {
        "code_2": "VG",
        "code_3": "VGB",
        "continent": "North America",
        "country": "British Virgin Islands",
        "num": "92",
    },
    "Brunei Darussalam": {
        "code_2": "BN",
        "code_3": "BRN",
        "continent": "Asia",
        "country": "Brunei Darussalam",
        "num": "96",
    },
    "Bulgaria": {
        "code_2": "BG",
        "code_3": "BGR",
        "continent": "Europe",
        "country": "Bulgaria",
        "num": "100",
    },
    "Burkina Faso": {
        "code_2": "BF",
        "code_3": "BFA",
        "continent": "Africa",
        "country": "Burkina Faso",
        "num": "854",
    },
    "Burundi": {
        "code_2": "BI",
        "code_3": "BDI",
        "continent": "Africa",
        "country": "Burundi",
        "num": "108",
    },
    "Cambodia": {
        "code_2": "KH",
        "code_3": "KHM",
        "continent": "Asia",
        "country": "Cambodia",
        "num": "116",
    },
    "Cameroon": {
        "code_2": "CM",
        "code_3": "CMR",
        "continent": "Africa",
        "country": "Cameroon",
        "num": "120",
    },
    "Canada": {
        "code_2": "CA",
        "code_3": "CAN",
        "continent": "North America",
        "country": "Canada",
        "num": "124",
    },
    "Cape Verde": {
        "code_2": "CV",
        "code_3": "CPV",
        "continent": "Africa",
        "country": "Cape Verde",
        "num": "132",
    },
    "Cayman Islands": {
        "code_2": "KY",
        "code_3": "CYM",
        "continent": "North America",
        "country": "Cayman Islands",
        "num": "136",
    },
    "Central African Republic": {
        "code_2": "CF",
        "code_3": "CAF",
        "continent": "Africa",
        "country": "Central African Republic",
        "num": "140",
    },
    "Chad": {
        "code_2": "TD",
        "code_3": "TCD",
        "continent": "Africa",
        "country": "Chad",
        "num": "148",
    },
    "Chagos Archipelago": {
        "code_2": "IO",
        "code_3": "IOT",
        "continent": "Asia",
        "country": "Chagos Archipelago",
        "num": "86",
    },
    "Chile": {
        "code_2": "CL",
        "code_3": "CHL",
        "continent": "South America",
        "country": "Chile",
        "num": "152",
    },
    "China": {
        "code_2": "CN",
        "code_3": "CHN",
        "continent": "Asia",
        "country": "China",
        "num": "156",
    },
    "Christmas Island": {
        "code_2": "CX",
        "code_3": "CXR",
        "continent": "Asia",
        "country": "Christmas Island",
        "num": "162",
    },
    "Cocos (Keeling) Islands": {
        "code_2": "CC",
        "code_3": "CCK",
        "continent": "Asia",
        "country": "Cocos (Keeling) Islands",
        "num": "166",
    },
    "Colombia": {
        "code_2": "CO",
        "code_3": "COL",
        "continent": "South America",
        "country": "Colombia",
        "num": "170",
    },
    "Comoros": {
        "code_2": "KM",
        "code_3": "COM",
        "continent": "Africa",
        "country": "Comoros",
        "num": "174",
    },
    "Congo": {
        "code_2": "CG",
        "code_3": "COG",
        "continent": "Africa",
        "country": "Congo",
        "num": "178",
    },
    "Cook Islands": {
        "code_2": "CK",
        "code_3": "COK",
        "continent": "Oceania",
        "country": "Cook Islands",
        "num": "184",
    },
    "Costa Rica": {
        "code_2": "CR",
        "code_3": "CRI",
        "continent": "North America",
        "country": "Costa Rica",
        "num": "188",
    },
    "Cote d'Ivoire": {
        "code_2": "CI",
        "code_3": "CIV",
        "continent": "Africa",
        "country": "Cote d'Ivoire",
        "num": "384",
    },
    "Croatia": {
        "code_2": "HR",
        "code_3": "HRV",
        "continent": "Europe",
        "country": "Croatia",
        "num": "191",
    },
    "Cuba": {
        "code_2": "CU",
        "code_3": "CUB",
        "continent": "North America",
        "country": "Cuba",
        "num": "192",
    },
    "Curaçao": {
        "code_2": "CW",
        "code_3": "CUW",
        "continent": "North America",
        "country": "Curaçao",
        "num": "531",
    },
    "Cyprus": {
        "code_2": "CY",
        "code_3": "CYP",
        "continent": "Europe",
        "country": "Cyprus",
        "num": "196",
    },
    "Czech Republic": {
        "code_2": "CZ",
        "code_3": "CZE",
        "continent": "Europe",
        "country": "Czech Republic",
        "num": "203",
    },
    "Democratic Republic of the Congo": {
        "code_2": "CD",
        "code_3": "COD",
        "continent": "Africa",
        "country": "Democratic Republic of the Congo",
        "num": "180",
    },
    "Denmark": {
        "code_2": "DK",
        "code_3": "DNK",
        "continent": "Europe",
        "country": "Denmark",
        "num": "208",
    },
    "Djibouti": {
        "code_2": "DJ",
        "code_3": "DJI",
        "continent": "Africa",
        "country": "Djibouti",
        "num": "262",
    },
    "Dominica": {
        "code_2": "DM",
        "code_3": "DMA",
        "continent": "North America",
        "country": "Dominica",
        "num": "212",
    },
    "Dominican Republic": {
        "code_2": "DO",
        "code_3": "DOM",
        "continent": "North America",
        "country": "Dominican Republic",
        "num": "214",
    },
    "Ecuador": {
        "code_2": "EC",
        "code_3": "ECU",
        "continent": "South America",
        "country": "Ecuador",
        "num": "218",
    },
    "Egypt": {
        "code_2": "EG",
        "code_3": "EGY",
        "continent": "Africa",
        "country": "Egypt",
        "num": "818",
    },
    "El Salvador": {
        "code_2": "SV",
        "code_3": "SLV",
        "continent": "North America",
        "country": "El Salvador",
        "num": "222",
    },
    "Equatorial Guinea": {
        "code_2": "GQ",
        "code_3": "GNQ",
        "continent": "Africa",
        "country": "Equatorial Guinea",
        "num": "226",
    },
    "Eritrea": {
        "code_2": "ER",
        "code_3": "ERI",
        "continent": "Africa",
        "country": "Eritrea",
        "num": "232",
    },
    "Estonia": {
        "code_2": "EE",
        "code_3": "EST",
        "continent": "Europe",
        "country": "Estonia",
        "num": "233",
    },
    "Ethiopia": {
        "code_2": "ET",
        "code_3": "ETH",
        "continent": "Africa",
        "country": "Ethiopia",
        "num": "231",
    },
    "Falkland Islands (Malvinas)": {
        "code_2": "FK",
        "code_3": "FLK",
        "continent": "South America",
        "country": "Falkland Islands (Malvinas)",
        "num": "238",
    },
    "Faroe Islands": {
        "code_2": "FO",
        "code_3": "FRO",
        "continent": "Europe",
        "country": "Faroe Islands",
        "num": "234",
    },
    "Fiji": {
        "code_2": "FJ",
        "code_3": "FJI",
        "continent": "Oceania",
        "country": "Fiji",
        "num": "242",
    },
    "Finland": {
        "code_2": "FI",
        "code_3": "FIN",
        "continent": "Europe",
        "country": "Finland",
        "num": "246",
    },
    "France": {
        "code_2": "FR",
        "code_3": "FRA",
        "continent": "Europe",
        "country": "France",
        "num": "250",
    },
    "French Guiana": {
        "code_2": "GF",
        "code_3": "GUF",
        "continent": "South America",
        "country": "French Guiana",
        "num": "254",
    },
    "French Polynesia": {
        "code_2": "PF",
        "code_3": "PYF",
        "continent": "Oceania",
        "country": "French Polynesia",
        "num": "258",
    },
    "French Southern Territories": {
        "code_2": "TF",
        "code_3": "ATF",
        "continent": "Antarctica",
        "country": "French Southern Territories",
        "num": "260",
    },
    "Gabon": {
        "code_2": "GA",
        "code_3": "GAB",
        "continent": "Africa",
        "country": "Gabon",
        "num": "266",
    },
    "Gambia the": {
        "code_2": "GM",
        "code_3": "GMB",
        "continent": "Africa",
        "country": "Gambia the",
        "num": "270",
    },
    "Georgia": {
        "code_2": "GE",
        "code_3": "GEO",
        "continent": "Europe",
        "country": "Georgia",
        "num": "268",
    },
    "Germany": {
        "code_2": "DE",
        "code_3": "DEU",
        "continent": "Europe",
        "country": "Germany",
        "num": "276",
    },
    "Ghana": {
        "code_2": "GH",
        "code_3": "GHA",
        "continent": "Africa",
        "country": "Ghana",
        "num": "288",
    },
    "Gibraltar": {
        "code_2": "GI",
        "code_3": "GIB",
        "continent": "Europe",
        "country": "Gibraltar",
        "num": "292",
    },
    "Greece": {
        "code_2": "GR",
        "code_3": "GRC",
        "continent": "Europe",
        "country": "Greece",
        "num": "300",
    },
    "Greenland": {
        "code_2": "GL",
        "code_3": "GRL",
        "continent": "North America",
        "country": "Greenland",
        "num": "304",
    },
    "Grenada": {
        "code_2": "GD",
        "code_3": "GRD",
        "continent": "North America",
        "country": "Grenada",
        "num": "308",
    },
    "Guadeloupe": {
        "code_2": "GP",
        "code_3": "GLP",
        "continent": "North America",
        "country": "Guadeloupe",
        "num": "312",
    },
    "Guam": {
        "code_2": "GU",
        "code_3": "GUM",
        "continent": "Oceania",
        "country": "Guam",
        "num": "316",
    },
    "Guatemala": {
        "code_2": "GT",
        "code_3": "GTM",
        "continent": "North America",
        "country": "Guatemala",
        "num": "320",
    },
    "Guernsey": {
        "code_2": "GG",
        "code_3": "GGY",
        "continent": "Europe",
        "country": "Guernsey",
        "num": "831",
    },
    "Guinea": {
        "code_2": "GN",
        "code_3": "GIN",
        "continent": "Africa",
        "country": "Guinea",
        "num": "324",
    },
    "Guinea-Bissau": {
        "code_2": "GW",
        "code_3": "GNB",
        "continent": "Africa",
        "country": "Guinea-Bissau",
        "num": "624",
    },
    "Guyana": {
        "code_2": "GY",
        "code_3": "GUY",
        "continent": "South America",
        "country": "Guyana",
        "num": "328",
    },
    "Haiti": {
        "code_2": "HT",
        "code_3": "HTI",
        "continent": "North America",
        "country": "Haiti",
        "num": "332",
    },
    "Heard Island and McDonald Islands": {
        "code_2": "HM",
        "code_3": "HMD",
        "continent": "Antarctica",
        "country": "Heard Island and McDonald Islands",
        "num": "334",
    },
    "Holy See (Vatican City State)": {
        "code_2": "VA",
        "code_3": "VAT",
        "continent": "Europe",
        "country": "Holy See (Vatican City State)",
        "num": "336",
    },
    "Honduras": {
        "code_2": "HN",
        "code_3": "HND",
        "continent": "North America",
        "country": "Honduras",
        "num": "340",
    },
    "Hong Kong": {
        "code_2": "HK",
        "code_3": "HKG",
        "continent": "Asia",
        "country": "Hong Kong",
        "num": "344",
    },
    "Hungary": {
        "code_2": "HU",
        "code_3": "HUN",
        "continent": "Europe",
        "country": "Hungary",
        "num": "348",
    },
    "Iceland": {
        "code_2": "IS",
        "code_3": "ISL",
        "continent": "Europe",
        "country": "Iceland",
        "num": "352",
    },
    "India": {
        "code_2": "IN",
        "code_3": "IND",
        "continent": "Asia",
        "country": "India",
        "num": "356",
    },
    "Indonesia": {
        "code_2": "ID",
        "code_3": "IDN",
        "continent": "Asia",
        "country": "Indonesia",
        "num": "360",
    },
    "Iran": {
        "code_2": "IR",
        "code_3": "IRN",
        "continent": "Asia",
        "country": "Iran",
        "num": "364",
    },
    "Iraq": {
        "code_2": "IQ",
        "code_3": "IRQ",
        "continent": "Asia",
        "country": "Iraq",
        "num": "368",
    },
    "Ireland": {
        "code_2": "IE",
        "code_3": "IRL",
        "continent": "Europe",
        "country": "Ireland",
        "num": "372",
    },
    "Isle of Man": {
        "code_2": "IM",
        "code_3": "IMN",
        "continent": "Europe",
        "country": "Isle of Man",
        "num": "833",
    },
    "Israel": {
        "code_2": "IL",
        "code_3": "ISR",
        "continent": "Asia",
        "country": "Israel",
        "num": "376",
    },
    "Italy": {
        "code_2": "IT",
        "code_3": "ITA",
        "continent": "Europe",
        "country": "Italy",
        "num": "380",
    },
    "Jamaica": {
        "code_2": "JM",
        "code_3": "JAM",
        "continent": "North America",
        "country": "Jamaica",
        "num": "388",
    },
    "Japan": {
        "code_2": "JP",
        "code_3": "JPN",
        "continent": "Asia",
        "country": "Japan",
        "num": "392",
    },
    "Jersey": {
        "code_2": "JE",
        "code_3": "JEY",
        "continent": "Europe",
        "country": "Jersey",
        "num": "832",
    },
    "Jordan": {
        "code_2": "JO",
        "code_3": "JOR",
        "continent": "Asia",
        "country": "Jordan",
        "num": "400",
    },
    "Kazakhstan": {
        "code_2": "KZ",
        "code_3": "KAZ",
        "continent": "Europe",
        "country": "Kazakhstan",
        "num": "398",
    },
    "Kenya": {
        "code_2": "KE",
        "code_3": "KEN",
        "continent": "Africa",
        "country": "Kenya",
        "num": "404",
    },
    "Kiribati": {
        "code_2": "KI",
        "code_3": "KIR",
        "continent": "Oceania",
        "country": "Kiribati",
        "num": "296",
    },
    "Korea": {
        "code_2": "KR",
        "code_3": "KOR",
        "continent": "Asia",
        "country": "Korea",
        "num": "410",
    },
    "Kuwait": {
        "code_2": "KW",
        "code_3": "KWT",
        "continent": "Asia",
        "country": "Kuwait",
        "num": "414",
    },
    "Kyrgyz Republic": {
        "code_2": "KG",
        "code_3": "KGZ",
        "continent": "Asia",
        "country": "Kyrgyz Republic",
        "num": "417",
    },
    "Lao": {
        "code_2": "LA",
        "code_3": "LAO",
        "continent": "Asia",
        "country": "Lao",
        "num": "418",
    },
    "Latvia": {
        "code_2": "LV",
        "code_3": "LVA",
        "continent": "Europe",
        "country": "Latvia",
        "num": "428",
    },
    "Lebanon": {
        "code_2": "LB",
        "code_3": "LBN",
        "continent": "Asia",
        "country": "Lebanon",
        "num": "422",
    },
    "Lesotho": {
        "code_2": "LS",
        "code_3": "LSO",
        "continent": "Africa",
        "country": "Lesotho",
        "num": "426",
    },
    "Liberia": {
        "code_2": "LR",
        "code_3": "LBR",
        "continent": "Africa",
        "country": "Liberia",
        "num": "430",
    },
    "Libyan Arab Jamahiriya": {
        "code_2": "LY",
        "code_3": "LBY",
        "continent": "Africa",
        "country": "Libyan Arab Jamahiriya",
        "num": "434",
    },
    "Liechtenstein": {
        "code_2": "LI",
        "code_3": "LIE",
        "continent": "Europe",
        "country": "Liechtenstein",
        "num": "438",
    },
    "Lithuania": {
        "code_2": "LT",
        "code_3": "LTU",
        "continent": "Europe",
        "country": "Lithuania",
        "num": "440",
    },
    "Luxembourg": {
        "code_2": "LU",
        "code_3": "LUX",
        "continent": "Europe",
        "country": "Luxembourg",
        "num": "442",
    },
    "Macao": {
        "code_2": "MO",
        "code_3": "MAC",
        "continent": "Asia",
        "country": "Macao",
        "num": "446",
    },
    "Macedonia": {
        "code_2": "MK",
        "code_3": "MKD",
        "continent": "Europe",
        "country": "Macedonia",
        "num": "807",
    },
    "Madagascar": {
        "code_2": "MG",
        "code_3": "MDG",
        "continent": "Africa",
        "country": "Madagascar",
        "num": "450",
    },
    "Malawi": {
        "code_2": "MW",
        "code_3": "MWI",
        "continent": "Africa",
        "country": "Malawi",
        "num": "454",
    },
    "Malaysia": {
        "code_2": "MY",
        "code_3": "MYS",
        "continent": "Asia",
        "country": "Malaysia",
        "num": "458",
    },
    "Maldives": {
        "code_2": "MV",
        "code_3": "MDV",
        "continent": "Asia",
        "country": "Maldives",
        "num": "462",
    },
    "Mali": {
        "code_2": "ML",
        "code_3": "MLI",
        "continent": "Africa",
        "country": "Mali",
        "num": "466",
    },
    "Malta": {
        "code_2": "MT",
        "code_3": "MLT",
        "continent": "Europe",
        "country": "Malta",
        "num": "470",
    },
    "Marshall Islands the": {
        "code_2": "MH",
        "code_3": "MHL",
        "continent": "Oceania",
        "country": "Marshall Islands the",
        "num": "584",
    },
    "Martinique": {
        "code_2": "MQ",
        "code_3": "MTQ",
        "continent": "North America",
        "country": "Martinique",
        "num": "474",
    },
    "Mauritania": {
        "code_2": "MR",
        "code_3": "MRT",
        "continent": "Africa",
        "country": "Mauritania",
        "num": "478",
    },
    "Mauritius": {
        "code_2": "MU",
        "code_3": "MUS",
        "continent": "Africa",
        "country": "Mauritius",
        "num": "480",
    },
    "Mayotte": {
        "code_2": "YT",
        "code_3": "MYT",
        "continent": "Africa",
        "country": "Mayotte",
        "num": "175",
    },
    "Mexico": {
        "code_2": "MX",
        "code_3": "MEX",
        "continent": "North America",
        "country": "Mexico",
        "num": "484",
    },
    "Micronesia": {
        "code_2": "FM",
        "code_3": "FSM",
        "continent": "Oceania",
        "country": "Micronesia",
        "num": "583",
    },
    "Moldova": {
        "code_2": "MD",
        "code_3": "MDA",
        "continent": "Europe",
        "country": "Moldova",
        "num": "498",
    },
    "Monaco": {
        "code_2": "MC",
        "code_3": "MCO",
        "continent": "Europe",
        "country": "Monaco",
        "num": "492",
    },
    "Mongolia": {
        "code_2": "MN",
        "code_3": "MNG",
        "continent": "Asia",
        "country": "Mongolia",
        "num": "496",
    },
    "Montenegro": {
        "code_2": "ME",
        "code_3": "MNE",
        "continent": "Europe",
        "country": "Montenegro",
        "num": "499",
    },
    "Montserrat": {
        "code_2": "MS",
        "code_3": "MSR",
        "continent": "North America",
        "country": "Montserrat",
        "num": "500",
    },
    "Morocco": {
        "code_2": "MA",
        "code_3": "MAR",
        "continent": "Africa",
        "country": "Morocco",
        "num": "504",
    },
    "Mozambique": {
        "code_2": "MZ",
        "code_3": "MOZ",
        "continent": "Africa",
        "country": "Mozambique",
        "num": "508",
    },
    "Myanmar": {
        "code_2": "MM",
        "code_3": "MMR",
        "continent": "Asia",
        "country": "Myanmar",
        "num": "104",
    },
    "Namibia": {
        "code_2": "NA",
        "code_3": "NAM",
        "continent": "Africa",
        "country": "Namibia",
        "num": "516",
    },
    "Nauru": {
        "code_2": "NR",
        "code_3": "NRU",
        "continent": "Oceania",
        "country": "Nauru",
        "num": "520",
    },
    "Nepal": {
        "code_2": "NP",
        "code_3": "NPL",
        "continent": "Asia",
        "country": "Nepal",
        "num": "524",
    },
    "Netherlands Antilles": {
        "code_2": "AN",
        "code_3": "ANT",
        "continent": "North America",
        "country": "Netherlands Antilles",
        "num": "530",
    },
    "Netherlands the": {
        "code_2": "NL",
        "code_3": "NLD",
        "continent": "Europe",
        "country": "Netherlands the",
        "num": "528",
    },
    "New Caledonia": {
        "code_2": "NC",
        "code_3": "NCL",
        "continent": "Oceania",
        "country": "New Caledonia",
        "num": "540",
    },
    "New Zealand": {
        "code_2": "NZ",
        "code_3": "NZL",
        "continent": "Oceania",
        "country": "New Zealand",
        "num": "554",
    },
    "Nicaragua": {
        "code_2": "NI",
        "code_3": "NIC",
        "continent": "North America",
        "country": "Nicaragua",
        "num": "558",
    },
    "Niger": {
        "code_2": "NE",
        "code_3": "NER",
        "continent": "Africa",
        "country": "Niger",
        "num": "562",
    },
    "Nigeria": {
        "code_2": "NG",
        "code_3": "NGA",
        "continent": "Africa",
        "country": "Nigeria",
        "num": "566",
    },
    "Niue": {
        "code_2": "NU",
        "code_3": "NIU",
        "continent": "Oceania",
        "country": "Niue",
        "num": "570",
    },
    "Norfolk Island": {
        "code_2": "NF",
        "code_3": "NFK",
        "continent": "Oceania",
        "country": "Norfolk Island",
        "num": "574",
    },
    "Northern Mariana Islands": {
        "code_2": "MP",
        "code_3": "MNP",
        "continent": "Oceania",
        "country": "Northern Mariana Islands",
        "num": "580",
    },
    "Norway": {
        "code_2": "NO",
        "code_3": "NOR",
        "continent": "Europe",
        "country": "Norway",
        "num": "578",
    },
    "Oman": {
        "code_2": "OM",
        "code_3": "OMN",
        "continent": "Asia",
        "country": "Oman",
        "num": "512",
    },
    "Pakistan": {
        "code_2": "PK",
        "code_3": "PAK",
        "continent": "Asia",
        "country": "Pakistan",
        "num": "586",
    },
    "Palau": {
        "code_2": "PW",
        "code_3": "PLW",
        "continent": "Oceania",
        "country": "Palau",
        "num": "585",
    },
    "Palestinian": {
        "code_2": "PS",
        "code_3": "PSE",
        "continent": "Asia",
        "country": "Palestinian",
        "num": "275",
    },
    "Panama": {
        "code_2": "PA",
        "code_3": "PAN",
        "continent": "North America",
        "country": "Panama",
        "num": "591",
    },
    "Papua New Guinea": {
        "code_2": "PG",
        "code_3": "PNG",
        "continent": "Oceania",
        "country": "Papua New Guinea",
        "num": "598",
    },
    "Paraguay": {
        "code_2": "PY",
        "code_3": "PRY",
        "continent": "South America",
        "country": "Paraguay",
        "num": "600",
    },
    "Peru": {
        "code_2": "PE",
        "code_3": "PER",
        "continent": "South America",
        "country": "Peru",
        "num": "604",
    },
    "Philippines the": {
        "code_2": "PH",
        "code_3": "PHL",
        "continent": "Asia",
        "country": "Philippines the",
        "num": "608",
    },
    "Pitcairn Islands": {
        "code_2": "PN",
        "code_3": "PCN",
        "continent": "Oceania",
        "country": "Pitcairn Islands",
        "num": "612",
    },
    "Poland": {
        "code_2": "PL",
        "code_3": "POL",
        "continent": "Europe",
        "country": "Poland",
        "num": "616",
    },
    "Portugal": {
        "code_2": "PT",
        "code_3": "PRT",
        "continent": "Europe",
        "country": "Portugal",
        "num": "620",
    },
    "Puerto Rico": {
        "code_2": "PR",
        "code_3": "PRI",
        "continent": "North America",
        "country": "Puerto Rico",
        "num": "630",
    },
    "Qatar": {
        "code_2": "QA",
        "code_3": "QAT",
        "continent": "Asia",
        "country": "Qatar",
        "num": "634",
    },
    "Reunion": {
        "code_2": "RE",
        "code_3": "REU",
        "continent": "Africa",
        "country": "Reunion",
        "num": "638",
    },
    "Romania": {
        "code_2": "RO",
        "code_3": "ROU",
        "continent": "Europe",
        "country": "Romania",
        "num": "642",
    },
    "Russian Federation": {
        "code_2": "RU",
        "code_3": "RUS",
        "continent": "Europe",
        "country": "Russian Federation",
        "num": "643",
    },
    "Rwanda": {
        "code_2": "RW",
        "code_3": "RWA",
        "continent": "Africa",
        "country": "Rwanda",
        "num": "646",
    },
    "Saint Barthelemy": {
        "code_2": "BL",
        "code_3": "BLM",
        "continent": "North America",
        "country": "Saint Barthelemy",
        "num": "652",
    },
    "Saint Helena": {
        "code_2": "SH",
        "code_3": "SHN",
        "continent": "Africa",
        "country": "Saint Helena",
        "num": "654",
    },
    "Saint Kitts and Nevis": {
        "code_2": "KN",
        "code_3": "KNA",
        "continent": "North America",
        "country": "Saint Kitts and Nevis",
        "num": "659",
    },
    "Saint Lucia": {
        "code_2": "LC",
        "code_3": "LCA",
        "continent": "North America",
        "country": "Saint Lucia",
        "num": "662",
    },
    "Saint Martin": {
        "code_2": "MF",
        "code_3": "MAF",
        "continent": "North America",
        "country": "Saint Martin",
        "num": "663",
    },
    "Saint Pierre and Miquelon": {
        "code_2": "PM",
        "code_3": "SPM",
        "continent": "North America",
        "country": "Saint Pierre and Miquelon",
        "num": "666",
    },
    "Saint Vincent and the Grenadines": {
        "code_2": "VC",
        "code_3": "VCT",
        "continent": "North America",
        "country": "Saint Vincent and the   Grenadines",
        "num": "670",
    },
    "Samoa": {
        "code_2": "WS",
        "code_3": "WSM",
        "continent": "Oceania",
        "country": "Samoa",
        "num": "882",
    },
    "San Marino": {
        "code_2": "SM",
        "code_3": "SMR",
        "continent": "Europe",
        "country": "San Marino",
        "num": "674",
    },
    "Sao Tome and Principe": {
        "code_2": "ST",
        "code_3": "STP",
        "continent": "Africa",
        "country": "Sao Tome and Principe",
        "num": "678",
    },
    "Saudi Arabia": {
        "code_2": "SA",
        "code_3": "SAU",
        "continent": "Asia",
        "country": "Saudi Arabia",
        "num": "682",
    },
    "Senegal": {
        "code_2": "SN",
        "code_3": "SEN",
        "continent": "Africa",
        "country": "Senegal",
        "num": "686",
    },
    "Serbia": {
        "code_2": "RS",
        "code_3": "SRB",
        "continent": "Europe",
        "country": "Serbia",
        "num": "688",
    },
    "Seychelles": {
        "code_2": "SC",
        "code_3": "SYC",
        "continent": "Africa",
        "country": "Seychelles",
        "num": "690",
    },
    "Sierra Leone": {
        "code_2": "SL",
        "code_3": "SLE",
        "continent": "Africa",
        "country": "Sierra Leone",
        "num": "694",
    },
    "Singapore": {
        "code_2": "SG",
        "code_3": "SGP",
        "continent": "Asia",
        "country": "Singapore",
        "num": "702",
    },
    "Sint Maarten (Netherlands)": {
        "code_2": "SX",
        "code_3": "SXM",
        "continent": "North America",
        "country": "Sint Maarten (Netherlands)",
        "num": "534",
    },
    "Slovakia": {
        "code_2": "SK",
        "code_3": "SVK",
        "continent": "Europe",
        "country": "Slovakia",
        "num": "703",
    },
    "Slovenia": {
        "code_2": "SI",
        "code_3": "SVN",
        "continent": "Europe",
        "country": "Slovenia",
        "num": "705",
    },
    "Solomon Islands": {
        "code_2": "SB",
        "code_3": "SLB",
        "continent": "Oceania",
        "country": "Solomon Islands",
        "num": "90",
    },
    "Somalia": {
        "code_2": "SO",
        "code_3": "SOM",
        "continent": "Africa",
        "country": "Somalia",
        "num": "706",
    },
    "South Africa": {
        "code_2": "ZA",
        "code_3": "ZAF",
        "continent": "Africa",
        "country": "South Africa",
        "num": "710",
    },
    "South Georgia and the South Sandwich Islands": {
        "code_2": "GS",
        "code_3": "SGS",
        "continent": "Antarctica",
        "country": "South Georgia and the South Sandwich Islands",
        "num": "239",
    },
    "South Sudan": {
        "code_2": "SS",
        "code_3": "SSD",
        "continent": "Africa",
        "country": "South Sudan",
        "num": "728",
    },
    "Spain": {
        "code_2": "ES",
        "code_3": "ESP",
        "continent": "Europe",
        "country": "Spain",
        "num": "724",
    },
    "Spratly Islands": {
        "code_2": "XS",
        "code_3": "null",
        "continent": "Asia",
        "country": "Spratly Islands",
        "num": "null",
    },
    "Sri Lanka": {
        "code_2": "LK",
        "code_3": "LKA",
        "continent": "Asia",
        "country": "Sri Lanka",
        "num": "144",
    },
    "Sudan": {
        "code_2": "SD",
        "code_3": "SDN",
        "continent": "Africa",
        "country": "Sudan",
        "num": "736",
    },
    "Suricountry": {
        "code_2": "SR",
        "code_3": "SUR",
        "continent": "South America",
        "country": "Suricountry",
        "num": "740",
    },
    "Svalbard & Jan Mayen Islands": {
        "code_2": "SJ",
        "code_3": "SJM",
        "continent": "Europe",
        "country": "Svalbard & Jan Mayen Islands",
        "num": "744",
    },
    "Swaziland": {
        "code_2": "SZ",
        "code_3": "SWZ",
        "continent": "Africa",
        "country": "Swaziland",
        "num": "748",
    },
    "Sweden": {
        "code_2": "SE",
        "code_3": "SWE",
        "continent": "Europe",
        "country": "Sweden",
        "num": "752",
    },
    "Switzerland": {
        "code_2": "CH",
        "code_3": "CHE",
        "continent": "Europe",
        "country": "Switzerland",
        "num": "756",
    },
    "Syrian Arab Republic": {
        "code_2": "SY",
        "code_3": "SYR",
        "continent": "Asia",
        "country": "Syrian Arab Republic",
        "num": "760",
    },
    "Taiwan": {
        "code_2": "TW",
        "code_3": "TWN",
        "continent": "Asia",
        "country": "Taiwan",
        "num": "158",
    },
    "Tajikistan": {
        "code_2": "TJ",
        "code_3": "TJK",
        "continent": "Asia",
        "country": "Tajikistan",
        "num": "762",
    },
    "Tanzania": {
        "code_2": "TZ",
        "code_3": "TZA",
        "continent": "Africa",
        "country": "Tanzania",
        "num": "834",
    },
    "Thailand": {
        "code_2": "TH",
        "code_3": "THA",
        "continent": "Asia",
        "country": "Thailand",
        "num": "764",
    },
    "Timor-Leste": {
        "code_2": "TL",
        "code_3": "TLS",
        "continent": "Asia",
        "country": "Timor-Leste",
        "num": "626",
    },
    "Togo": {
        "code_2": "TG",
        "code_3": "TGO",
        "continent": "Africa",
        "country": "Togo",
        "num": "768",
    },
    "Tokelau": {
        "code_2": "TK",
        "code_3": "TKL",
        "continent": "Oceania",
        "country": "Tokelau",
        "num": "772",
    },
    "Tonga": {
        "code_2": "TO",
        "code_3": "TON",
        "continent": "Oceania",
        "country": "Tonga",
        "num": "776",
    },
    "Trinidad and Tobago": {
        "code_2": "TT",
        "code_3": "TTO",
        "continent": "North America",
        "country": "Trinidad and Tobago",
        "num": "780",
    },
    "Tunisia": {
        "code_2": "TN",
        "code_3": "TUN",
        "continent": "Africa",
        "country": "Tunisia",
        "num": "788",
    },
    "Turkey": {
        "code_2": "TR",
        "code_3": "TUR",
        "continent": "Europe",
        "country": "Turkey",
        "num": "792",
    },
    "Turkmenistan": {
        "code_2": "TM",
        "code_3": "TKM",
        "continent": "Asia",
        "country": "Turkmenistan",
        "num": "795",
    },
    "Turks and Caicos Islands": {
        "code_2": "TC",
        "code_3": "TCA",
        "continent": "North America",
        "country": "Turks and Caicos Islands",
        "num": "796",
    },
    "Tuvalu": {
        "code_2": "TV",
        "code_3": "TUV",
        "continent": "Oceania",
        "country": "Tuvalu",
        "num": "798",
    },
    "Uganda": {
        "code_2": "UG",
        "code_3": "UGA",
        "continent": "Africa",
        "country": "Uganda",
        "num": "800",
    },
    "Ukraine": {
        "code_2": "UA",
        "code_3": "UKR",
        "continent": "Europe",
        "country": "Ukraine",
        "num": "804",
    },
    "United Arab Emirates": {
        "code_2": "AE",
        "code_3": "ARE",
        "continent": "Asia",
        "country": "United Arab Emirates",
        "num": "784",
    },
    "United Kingdom": {
        "code_2": "GB",
        "code_3": "GBR",
        "continent": "Europe",
        "country": "United Kingdom",
        "num": "826",
    },
    "United States Minor Outlying Islands": {
        "code_2": "UM",
        "code_3": "UMI",
        "continent": "Oceania",
        "country": "United States Minor Outlying Islands",
        "num": "581",
    },
    "United States Virgin Islands": {
        "code_2": "VI",
        "code_3": "VIR",
        "continent": "North America",
        "country": "United States Virgin Islands",
        "num": "850",
    },
    "United States of America": {
        "code_2": "US",
        "code_3": "USA",
        "continent": "North America",
        "country": "United States of America",
        "num": "840",
    },
    "Uruguay": {
        "code_2": "UY",
        "code_3": "URY",
        "continent": "South America",
        "country": "Uruguay",
        "num": "858",
    },
    "Uzbekistan": {
        "code_2": "UZ",
        "code_3": "UZB",
        "continent": "Asia",
        "country": "Uzbekistan",
        "num": "860",
    },
    "Vanuatu": {
        "code_2": "VU",
        "code_3": "VUT",
        "continent": "Oceania",
        "country": "Vanuatu",
        "num": "548",
    },
    "Venezuela": {
        "code_2": "VE",
        "code_3": "VEN",
        "continent": "South America",
        "country": "Venezuela",
        "num": "862",
    },
    "Vietnam": {
        "code_2": "VN",
        "code_3": "VNM",
        "continent": "Asia",
        "country": "Vietnam",
        "num": "704",
    },
    "Wallis and Futuna": {
        "code_2": "WF",
        "code_3": "WLF",
        "continent": "Oceania",
        "country": "Wallis and Futuna",
        "num": "876",
    },
    "Western Sahara": {
        "code_2": "EH",
        "code_3": "ESH",
        "continent": "Africa",
        "country": "Western Sahara",
        "num": "732",
    },
    "Yemen": {
        "code_2": "YE",
        "code_3": "YEM",
        "continent": "Asia",
        "country": "Yemen",
        "num": "887",
    },
    "Zambia": {
        "code_2": "ZM",
        "code_3": "ZMB",
        "continent": "Africa",
        "country": "Zambia",
        "num": "894",
    },
    "Zimbabwe": {
        "code_2": "ZW",
        "code_3": "ZWE",
        "continent": "Africa",
        "country": "Zimbabwe",
        "num": "716",
    },
    "Åland Islands": {
        "code_2": "AX",
        "code_3": "ALA",
        "continent": "Europe",
        "country": "Åland Islands",
        "num": "248",
    },
}



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab

from pprint import pprint
from functools import partial

from riko import get_path
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

BR = {"find": "<br>"}
DEF_CUR_CODE = "USD"

odesk_conf = {"url": get_path("odesk.json"), "path": "items"}
guru_conf = {"url": get_path("guru.json"), "path": "items"}
elance_conf = {"url": get_path("elance.json"), "path": "items"}
freelancer_conf = {"url": get_path("freelancer.json"), "path": "items"}


def make_regex(field, match, replace, default=None):
    result = {"field": field, "match": match, "replace": replace, "default": default}

    return result


def make_simplemath(other, op):
    return {"other": {"subkey": other, "type": "number"}, "op": op}


def add_source(source):
    subelement_conf = {"path": "k:source.content.1", "token_key": None}

    sourced = source.urlparse(field="link", assign="k:source").subelement(
        conf=subelement_conf, emit=False, assign="k:source"
    )

    return sourced


def add_id(source, rule, field="link"):
    make_id_part = [{"subkey": "k:source"}, {"value": "-"}, {"subkey": "id"}]

    ideed = source.strfind(conf={"rule": rule}, field=field, assign="id").strconcat(
        conf={"part": make_id_part}, assign="id"
    )

    return ideed


def add_posted(source, rule="", field="summary"):
    if rule:
        conf = {"rule": rule}
        source = source.strfind(conf=conf, field=field, assign="k:posted")
    else:
        rule = {"field": "updated", "newval": "k:posted"}
        source = source.rename(conf={"rule": rule})

    return source


def add_tags(source, rule, field="summary", assign="k:tags"):
    tokenizer_conf = {"dedupe": True, "sort": True}
    no_tags = {"field": assign}

    tag_strreplace_rule = [
        {"find": "  ", "replace": ","},
        {"find": "&gt;", "replace": ","},
        {"find": "&amp;", "replace": "&"},
        {"find": "Other -", "replace": ""},
        # {'find': '-', 'replace': ''},
    ]

    tagged = (
        source.strfind(conf={"rule": rule}, field=field, assign=assign)
        .strreplace(
            conf={"rule": tag_strreplace_rule},
            field=assign,
            assign=assign,
            skip_if=no_tags,
        )
        .strtransform(
            conf={"rule": {"transform": "lower"}},
            field=assign,
            assign=assign,
            skip_if=no_tags,
        )
        .tokenizer(conf=tokenizer_conf, field=assign, assign=assign, skip_if=no_tags)
    )

    return tagged


def add_budget(source, budget_text, fixed_text="", hourly_text="", double=True):
    codes = "$£€₹"
    no_raw_budget = {"field": "k:budget_raw"}
    has_code = {"field": "k:cur_code", "include": True}
    is_def_cur = {"field": "k:cur_code", "text": DEF_CUR_CODE}
    not_def_cur = {"field": "k:cur_code", "text": DEF_CUR_CODE, "include": True}
    isnt_fixed = {"field": "summary", "text": fixed_text, "include": True}
    isnt_hourly = {"field": "summary", "text": hourly_text, "include": True}
    no_symbol = {
        "field": "k:budget_raw",
        "text": codes,
        "op": "intersection",
        "include": True,
    }
    code_or_no_raw_budget = [has_code, no_raw_budget]
    def_cur_or_no_raw_budget = [is_def_cur, no_raw_budget]
    not_def_cur_or_no_raw_budget = [not_def_cur, no_raw_budget]

    first_num_rule = {"find": r"\d+", "location": "at"}
    last_num_rule = {"find": r"\d+", "location": "at", "param": "last"}
    cur_rule = {"find": r"\b[A-Z]{3}\b", "location": "at"}
    sym_rule = {"find": "[%s]" % codes, "location": "at"}

    # make_regex('k:budget_raw', r'[(),.\s]', ''),
    invalid_budgets = [
        {"find": "Less than", "replace": "0-"},
        {"find": "Under", "replace": "0-"},
        {"find": "Upto", "replace": "0-"},
        {"find": "or less", "replace": "-0"},
        {"find": "k", "replace": "000"},
        {"find": "Not Sure", "replace": ""},
        {"find": "Not sure", "replace": ""},
        {"find": "(", "replace": ""},
        {"find": ")", "replace": ""},
        {"find": ".", "replace": ""},
        {"find": ",", "replace": ""},
        {"find": " ", "replace": ""},
    ]

    cur_strreplace_rule = [
        {"find": "$", "replace": "USD"},
        {"find": "£", "replace": "GBP"},
        {"find": "€", "replace": "EUR"},
        {"find": "₹", "replace": "INR"},
    ]

    converted_budget_part = [
        {"subkey": "k:budget_w_sym"},
        {"value": " ("},
        {"subkey": "k:budget_converted_w_sym"},
        {"value": ")"},
    ]

    def_full_budget_part = {"subkey": "k:budget_w_sym"}
    hourly_budget_part = [{"subkey": "k:budget_full"}, {"value": " / hr"}]
    exchangerate_conf = {"url": get_path("quote.json")}
    native_currencyformat_conf = {"currency": {"subkey": "k:cur_code"}}
    def_currencyformat_conf = {"currency": DEF_CUR_CODE}
    ave_budget_conf = make_simplemath("k:budget_raw2_num", "mean")
    convert_budget_conf = make_simplemath("k:rate", "multiply")

    if fixed_text:
        source = source.strconcat(
            conf={"part": {"value": "fixed"}}, assign="k:job_type", skip_if=isnt_fixed
        )

    if hourly_text:
        source = source.strconcat(
            conf={"part": {"value": "hourly"}}, assign="k:job_type", skip_if=isnt_hourly
        )

    source = source.refind(
        conf={"rule": cur_rule},
        field="k:budget_raw",
        assign="k:cur_code",
        skip_if=no_raw_budget,
    ).strreplace(
        conf={"rule": invalid_budgets},
        field="k:budget_raw",
        assign="k:budget_raw",
        skip_if=no_raw_budget,
    )

    if double:
        source = (
            source.refind(
                conf={"rule": first_num_rule},
                field="k:budget_raw",
                assign="k:budget_raw_num",
                skip_if=no_raw_budget,
            )
            .refind(
                conf={"rule": last_num_rule},
                field="k:budget_raw",
                assign="k:budget_raw2_num",
                skip_if=no_raw_budget,
            )
            .simplemath(
                conf=ave_budget_conf,
                field="k:budget_raw_num",
                assign="k:budget",
                skip_if=no_raw_budget,
            )
        )
    else:
        source = source.refind(
            conf={"rule": first_num_rule},
            field="k:budget_raw",
            assign="k:budget",
            skip_if=no_raw_budget,
        )

    source = (
        source.refind(
            conf={"rule": sym_rule},
            field="k:budget_raw",
            assign="k:budget_raw_sym",
            skip_if=no_symbol,
        )
        .strreplace(
            conf={"rule": cur_strreplace_rule},
            field="k:budget_raw_sym",
            assign="k:cur_code",
            skip_if=code_or_no_raw_budget,
        )
        .currencyformat(
            conf=native_currencyformat_conf,
            field="k:budget",
            assign="k:budget_w_sym",
            skip_if=no_raw_budget,
        )
        .exchangerate(
            conf=exchangerate_conf,
            field="k:cur_code",
            assign="k:rate",
            skip_if=def_cur_or_no_raw_budget,
        )
        .simplemath(
            conf=convert_budget_conf,
            field="k:budget",
            assign="k:budget_converted",
            skip_if=def_cur_or_no_raw_budget,
        )
        .currencyformat(
            conf=def_currencyformat_conf,
            field="k:budget_converted",
            assign="k:budget_converted_w_sym",
            skip_if=def_cur_or_no_raw_budget,
        )
        .strconcat(
            conf={"part": converted_budget_part},
            assign="k:budget_full",
            skip_if=def_cur_or_no_raw_budget,
        )
        .strconcat(
            conf={"part": def_full_budget_part},
            assign="k:budget_full",
            skip_if=not_def_cur_or_no_raw_budget,
        )
    )

    if hourly_text:
        source = source.strconcat(
            conf={"part": hourly_budget_part},
            assign="k:budget_full",
            skip_if=isnt_hourly,
        )

    return source


def clean_locations(source):
    no_client_loc = {"field": "k:client_location"}
    no_work_loc = {"field": "k:work_location"}

    rule = {"find": ", ", "replace": ""}
    cleaned = source.strreplace(
        conf={"rule": rule},
        field="k:client_location",
        assign="k:client_location",
        skip_if=no_client_loc,
    ).strreplace(
        conf={"rule": rule},
        field="k:work_location",
        assign="k:work_location",
        skip_if=no_work_loc,
    )

    return cleaned


def remove_cruft(source):
    remove_rule = [
        {"field": "author"},
        {"field": "content"},
        {"field": "dc:creator"},
        {"field": "links"},
        {"field": "pubDate"},
        {"field": "summary"},
        {"field": "updated"},
        {"field": "updated_parsed"},
        {"field": "y:id"},
        {"field": "y:title"},
        {"field": "y:published"},
        {"field": "k:budget_raw"},
        {"field": "k:budget_raw2_num"},
        {"field": "k:budget_raw_num"},
        {"field": "k:budget_raw_sym"},
    ]

    return source.rename(conf={"rule": remove_rule})


def parse_odesk(source, stream=True):
    budget_text = "Budget</b>:"
    no_budget = {"field": "summary", "text": budget_text, "include": True}
    raw_budget_rule = [{"find": budget_text, "location": "after"}, BR]
    title_rule = {"find": "- oDesk"}
    find_id_rule = [{"find": "ID</b>:", "location": "after"}, BR]
    categ_rule = [{"find": "Category</b>:", "location": "after"}, BR]
    skills_rule = [{"find": "Skills</b>:", "location": "after"}, BR]
    client_loc_rule = [{"find": "Country</b>:", "location": "after"}, BR]
    posted_rule = [{"find": "Posted On</b>:", "location": "after"}, BR]
    desc_rule = [{"find": "<p>", "location": "after"}, {"find": "<br><br><b>"}]

    source = (
        source.strfind(conf={"rule": title_rule}, field="title", assign="title")
        .strfind(
            conf={"rule": client_loc_rule}, field="summary", assign="k:client_location"
        )
        .strfind(conf={"rule": desc_rule}, field="summary", assign="description")
        .strfind(
            conf={"rule": raw_budget_rule},
            field="summary",
            assign="k:budget_raw",
            skip_if=no_budget,
        )
    )

    source = add_source(source)
    source = add_posted(source, posted_rule)
    source = add_id(source, find_id_rule, field="summary")
    source = add_budget(source, budget_text, double=False)
    source = add_tags(source, skills_rule)
    source = add_tags(source, categ_rule, assign="k:categories")
    source = clean_locations(source)
    source = remove_cruft(source)
    return source.output if stream else source


def parse_guru(source, stream=True):
    budget_text = "budget:</b>"
    fixed_text = "Fixed Price budget:</b>"
    hourly_text = "Hourly budget:</b>"

    no_budget = {"field": "summary", "text": budget_text, "include": True}
    isnt_hourly = {"field": "summary", "text": hourly_text, "include": True}
    raw_budget_rule = [{"find": budget_text, "location": "after"}, BR]
    after_hourly = {"rule": {"find": "Rate:", "location": "after"}}
    find_id_rule = {"find": "/", "location": "after", "param": "last"}
    categ_rule = [{"find": "Category:</b>", "location": "after"}, BR]
    skills_rule = [{"find": "Required skills:</b>", "location": "after"}, BR]

    job_loc_conf = {
        "rule": [{"find": "Freelancer Location:</b>", "location": "after"}, BR]
    }

    desc_conf = {"rule": [{"find": "Description:</b>", "location": "after"}, BR]}

    source = (
        source.strfind(conf=job_loc_conf, field="summary", assign="k:work_location")
        .strfind(conf=desc_conf, field="summary", assign="description")
        .strfind(
            conf={"rule": raw_budget_rule},
            field="summary",
            assign="k:budget_raw",
            skip_if=no_budget,
        )
        .strfind(
            conf=after_hourly,
            field="k:budget_raw",
            assign="k:budget_raw",
            skip_if=isnt_hourly,
        )
    )

    kwargs = {"fixed_text": fixed_text, "hourly_text": hourly_text}
    source = add_source(source)
    source = add_posted(source)
    source = add_id(source, find_id_rule)
    source = add_budget(source, budget_text, **kwargs)
    source = add_tags(source, skills_rule)
    source = add_tags(source, categ_rule, assign="k:categories")
    source = clean_locations(source)
    source = remove_cruft(source)
    return source.output if stream else source


def parse_elance(source, stream=True):
    budget_text = "Budget:</b>"
    fixed_text = "Budget:</b> Fixed Price"
    hourly_text = "Budget:</b> Hourly"

    no_job_loc = {"field": "summary", "text": "Preferred Job Location", "include": True}
    no_client_loc = {"field": "summary", "text": "Client Location", "include": True}
    no_budget = {"field": "summary", "text": budget_text, "include": True}
    isnt_fixed = {"field": "summary", "text": fixed_text, "include": True}
    isnt_hourly = {"field": "summary", "text": hourly_text, "include": True}
    raw_budget_rule = [{"find": budget_text, "location": "after"}, BR]
    after_hourly = {"rule": {"find": "Hourly", "location": "after"}}
    after_fixed = {"rule": {"find": "Fixed Price", "location": "after"}}
    title_conf = {"rule": {"find": "| Elance Job"}}

    find_id_rule = [
        {"find": "/", "param": "last"},
        {"find": "/", "location": "after", "param": "last"},
    ]

    categ_rule = [{"find": "Category:</b>", "location": "after"}, BR]
    skills_rule = [{"find": "Desired Skills:</b>", "location": "after"}, BR]

    job_loc_conf = {
        "rule": [{"find": "Preferred Job Location:</b>", "location": "after"}, BR]
    }

    client_loc_conf = {
        "rule": [{"find": "Client Location:</b>", "location": "after"}, BR]
    }

    desc_rule = [{"find": "<p>", "location": "after"}, {"find": "...\n    <br>"}]

    proposals_conf = {
        "rule": [{"find": "Proposals:</b>", "location": "after"}, {"find": "("}]
    }

    jobs_posted_conf = {
        "rule": [
            {"find": "Client:</b> Client (", "location": "after"},
            {"find": "jobs posted"},
        ]
    }

    jobs_awarded_conf = {
        "rule": [{"find": "jobs posted,", "location": "after"}, {"find": "awarded"}]
    }

    purchased_conf = {
        "rule": [
            {"find": "total purchased"},
            {"find": ",", "location": "after", "param": "last"},
        ]
    }

    ends_conf = {
        "rule": [
            {"find": "Time Left:</b>", "location": "after"},
            {"find": ") <br>"},
            {"find": "h (Ends", "location": "after"},
        ]
    }

    source = (
        source.strfind(conf=title_conf, field="title", assign="title")
        .strfind(conf=proposals_conf, field="summary", assign="k:submissions")
        .strfind(conf=jobs_posted_conf, field="summary", assign="k:num_jobs")
        .strfind(conf=jobs_awarded_conf, field="summary", assign="k:per_awarded")
        .strfind(conf=purchased_conf, field="summary", assign="k:tot_purchased")
        .strfind(conf=ends_conf, field="summary", assign="k:due")
        .strfind(
            conf=job_loc_conf,
            field="summary",
            assign="k:work_location",
            skip_if=no_job_loc,
        )
        .strfind(
            conf=client_loc_conf,
            field="summary",
            assign="k:client_location",
            skip_if=no_client_loc,
        )
        .strfind(conf={"rule": desc_rule}, field="summary", assign="description")
        .strfind(
            conf={"rule": raw_budget_rule},
            field="summary",
            assign="k:budget_raw",
            skip_if=no_budget,
        )
        .strfind(
            conf=after_hourly,
            field="k:budget_raw",
            assign="k:budget_raw",
            skip_if=isnt_hourly,
        )
        .strfind(
            conf=after_fixed,
            field="k:budget_raw",
            assign="k:budget_raw",
            skip_if=isnt_fixed,
        )
    )

    kwargs = {"fixed_text": fixed_text, "hourly_text": hourly_text}
    source = add_source(source)
    source = add_posted(source)
    source = add_id(source, find_id_rule)
    source = add_budget(source, budget_text, **kwargs)
    source = add_tags(source, skills_rule)
    source = add_tags(source, categ_rule, assign="k:categories")
    source = clean_locations(source)
    source = remove_cruft(source)
    return source.output if stream else source


def parse_freelancer(source, stream=True):
    budget_text = "(Budget:"
    no_budget = {"field": "summary", "text": budget_text, "include": True}
    raw_budget_rule = [{"find": budget_text, "location": "after"}, {"find": ","}]

    title_rule = {"find": " by "}
    skills_rule = [{"find": ", Jobs:", "location": "after"}, {"find": ")</p>"}]
    desc_rule = [{"find": "<p>", "location": "after"}, {"find": "(Budget:"}]

    source = (
        source.strfind(conf={"rule": title_rule}, field="title", assign="title")
        .strfind(conf={"rule": desc_rule}, field="summary", assign="description")
        .strfind(
            conf={"rule": raw_budget_rule},
            field="summary",
            assign="k:budget_raw",
            skip_if=no_budget,
        )
    )

    source = add_source(source)
    source = add_posted(source)
    source = add_budget(source, budget_text)
    source = add_tags(source, skills_rule)
    source = clean_locations(source)
    source = remove_cruft(source)
    return source.output if stream else source


def pipe(test=False, parallel=False, threads=False):
    kwargs = {"parallel": parallel, "threads": threads}

    Pipe = partial(SyncPipe, "fetchdata", **kwargs)
    odesk_source = Pipe(conf=odesk_conf)
    guru_source = Pipe(conf=guru_conf)
    freelancer_source = Pipe(conf=freelancer_conf)
    elance_source = Pipe(conf=elance_conf)
    # odesk_source = SyncPipe('fetchdata', conf=odesk_conf, **kwargs)
    # guru_source = SyncPipe('fetchdata', conf=guru_conf, **kwargs)
    # elance_source = SyncPipe('fetchdata', conf=elance_conf, **kwargs)
    # freelancer_source = SyncPipe('fetchdata', conf=freelancer_conf, **kwargs)

    odesk_pipe = parse_odesk(odesk_source, stream=False)
    guru_stream = parse_guru(guru_source)
    elance_stream = parse_elance(elance_source)
    freelancer_stream = parse_freelancer(freelancer_source)

    others = [guru_stream, freelancer_stream, elance_stream]
    stream = odesk_pipe.union(others=others).list

    pprint(stream[-1])
    return stream


@coroutine
def async_pipe(reactor, test=None):
    Pipe = partial(AsyncPipe, "fetchdata")
    odesk_source = Pipe(conf=odesk_conf)
    guru_source = Pipe(conf=guru_conf)
    freelancer_source = Pipe(conf=freelancer_conf)
    elance_source = Pipe(conf=elance_conf)

    odesk_pipe = yield parse_odesk(odesk_source, stream=False)
    guru_stream = yield parse_guru(guru_source)
    elance_stream = yield parse_elance(elance_source)
    freelancer_stream = yield parse_freelancer(freelancer_source)

    others = [guru_stream, freelancer_stream, elance_stream]
    stream = odesk_pipe.union(others=others).list
    pprint(stream[-1])



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko
~~~~
Provides functions for analyzing and processing streams of structured data

Examples:
    basic usage::

        >>> from itertools import chain
        >>> from functools import partial
        >>> from riko.modules import itembuilder, strreplace
        >>> from riko.collections import SyncPipe
        >>>
        >>> ib_conf = {
        ...     'attrs': [
        ...         {'key': 'link', 'value': 'www.google.com', },
        ...         {'key': 'title', 'value': 'google', },
        ...         {'key': 'author', 'value': 'Tommy'}]}
        >>>
        >>> sr_conf = {
        ...     'rule': [{'find': 'Tom', 'param': 'first', 'replace': 'Tim'}]}
        >>>
        >>> items = itembuilder.pipe(conf=ib_conf)
        >>> pipe = partial(strreplace.pipe, conf=sr_conf, field='author')
        >>> replaced = map(pipe, items)
        >>> next(chain.from_iterable(replaced)) == {
        ...     'link': 'www.google.com', 'title': 'google',
        ...     'strreplace': 'Timmy', 'author': 'Tommy'}
        True
"""
from os import path as p

__version__ = "0.67.0"

__title__ = "riko"
__package_name__ = "riko"
__author__ = "Reuben Cummings"
__description__ = "A stream processing engine modeled after Yahoo! Pipes."
__email__ = "reubano@gmail.com"
__license__ = "MIT"
__copyright__ = "Copyright 2015 Reuben Cummings"

PARENT_DIR = p.abspath(p.dirname(__file__))
ENCODING = "utf-8"


def get_path(name):
    return "file://%s" % p.join(PARENT_DIR, "data", name)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.utils
~~~~~~~~~~~~~~
Provides utility classes and functions
"""
import re
import sys
import itertools as it
import fcntl

from math import isnan
from functools import partial, wraps
from operator import itemgetter
from os import O_NONBLOCK, path as p
from io import BytesIO, StringIO, TextIOBase
from urllib.error import HTTPError, URLError
from urllib.request import urlopen, Request

from werkzeug.local import LocalProxy

import requests
import pygogo as gogo
import mezmorize

try:
    import __builtin__ as _builtins
except ImportError:
    import builtins as _builtins

from meza import compat
from meza.io import reencode
from meza.fntools import SleepyDict
from mezmorize.utils import get_cache_type
from riko import ENCODING, __version__
from riko.cast import cast

logger = gogo.Gogo(__name__, verbose=False, monolog=True).logger

DEF_NS = "https://github.com/nerevu/riko"


def get_abspath(url):
    url = "http://%s" % url if url and "://" not in url else url

    if url and url.startswith("file:///"):
        # already have an abspath
        pass
    elif url and url.startswith("file://"):
        parent = p.dirname(p.dirname(__file__))
        rel_path = url[7:]
        abspath = p.abspath(p.join(parent, rel_path))
        url = "file://%s" % abspath

    return compat.decode(url)


# https://trac.edgewall.org/ticket/2066#comment:1
# http://stackoverflow.com/a/22675049/408556
def make_blocking(f):
    fd = f.fileno()
    flags = fcntl.fcntl(fd, fcntl.F_GETFL)

    if flags & O_NONBLOCK:
        blocking = flags & ~O_NONBLOCK
        fcntl.fcntl(fd, fcntl.F_SETFL, blocking)


if "nose" in sys.modules:
    logger.debug("Running in nose environment...")
    make_blocking(sys.stderr)


def default_user_agent(name="riko"):
    """
    Return a string representing the default user agent.
    :rtype: str
    """
    return "%s/%s" % (name, __version__)


class Chainable(object):
    def __init__(self, data, method=None):
        self.data = data
        self.method = method
        self.list = list(data)

    def __getattr__(self, name):
        funcs = (partial(getattr, x) for x in [self.data, _builtins, it])
        zipped = zip(funcs, it.repeat(AttributeError))
        method = multi_try(name, zipped, default=None)
        return Chainable(self.data, method)

    def __call__(self, *args, **kwargs):
        try:
            return Chainable(self.method(self.data, *args, **kwargs))
        except TypeError:
            return Chainable(self.method(args[0], self.data, **kwargs))


def invert_dict(d):
    return {v: k for k, v in d.items()}


def multi_try(source, zipped, default=None):
    value = None

    for func, error in zipped:
        try:
            value = func(source)
        except error:
            pass
        else:
            return value
    else:
        return default


def get_response_content_type(response):
    try:
        content_type = response.getheader("Content-Type", "")
    except AttributeError:
        content_type = response.headers.get("Content-Type", "")

    return content_type.lower()


def get_response_encoding(response, def_encoding=ENCODING):
    info = response.info()

    try:
        encoding = info.getencoding()
    except AttributeError:
        encoding = info.get_charset()

    encoding = None if encoding == "7bit" else encoding

    if not encoding and hasattr(info, "get_content_charset"):
        encoding = info.get_content_charset()

    if not encoding:
        content_type = get_response_content_type(response)

        if "charset" in content_type:
            ctype = content_type.split("=")[1]
            encoding = ctype.strip().strip('"').strip("'")

    extracted = encoding or def_encoding
    assert extracted
    return extracted


# https://docs.python.org/3.3/reference/expressions.html#examples
def auto_close(stream, f):
    try:
        for record in stream:
            yield record
    finally:
        f.close()


def opener(url, memoize=False, delay=0, encoding=ENCODING, params=None, **kwargs):
    params = params or {}
    timeout = kwargs.get("timeout")
    decode = kwargs.get("decode")

    if url.startswith("http") and params:
        r = requests.get(url, params=params, stream=True)
        r.raw.decode_content = decode
        response = r.text if memoize else r.raw
    else:
        req = Request(url, headers={"User-Agent": default_user_agent()})
        context = SleepyDict(delay=delay) if delay else None

        try:
            r = urlopen(req, context=context, timeout=timeout)
        except TypeError:
            r = urlopen(req, timeout=timeout)
        except HTTPError as e:
            raise URLError(f"{url} returned {e.code}: {e.reason}")
        except URLError as e:
            raise URLError(f"{url}: {e.reason}")

        text = r.read() if memoize else None

        if decode:
            encoding = get_response_encoding(r, encoding)

            if text:
                response = compat.decode(text, encoding)
            else:
                response = reencode(r.fp, encoding, decode=True)
                response.r = r
        else:
            response = text or r

    content_type = get_response_content_type(r)
    return (response, content_type)


def get_opener(memoize=False, **kwargs):
    wrapper = partial(opener, memoize=memoize, **kwargs)
    current_opener = wraps(opener)(wrapper)

    if memoize:
        kwargs.setdefault("cache_type", get_cache_type(spread=False))
        memoizer = mezmorize.memoize(**kwargs)
        current_opener = memoizer(current_opener)

    return current_opener


class fetch(TextIOBase):
    # http://stackoverflow.com/a/22836333/408556
    def __init__(self, url=None, memoize=False, **kwargs):
        # TODO: need to use separate timeouts for memoize and urlopen
        if memoize:
            self.opener = LocalProxy(lambda: get_opener(memoize=True, **kwargs))
        else:
            self.opener = get_opener(**kwargs)

        responses = self.opener(get_abspath(url))

        try:
            response, self.content_type = responses
        except ValueError:
            # HACK: This happens for memoized responses. Not sure why though!
            response, self.content_type = responses, "application/json"

        if memoize:
            wrapper = StringIO if kwargs.get("decode") else BytesIO
            f = wrapper(response)
        else:
            f = response

        self.close = f.close
        self.read = f.read
        self.readline = f.readline

        try:
            self.seek = f.seek
        except AttributeError:
            pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def ext(self):
        if not self.content_type:
            ext = None
        elif "xml" in self.content_type:
            ext = "xml"
        elif "json" in self.content_type:
            ext = "json"
        else:
            ext = self.content_type.split("/")[1].split(";")[0]

        return ext


def def_itemgetter(attr, default=0, _type=None):
    # like operator.itemgetter but fills in missing keys with a default value
    def keyfunc(item):
        value = item.get(attr, default)
        casted = cast(value, _type) if _type else value

        try:
            is_nan = isnan(casted)
        except TypeError:
            is_nan = False

        return default if is_nan else casted

    return keyfunc


# TODO: move this to meza.process.group
def group_by(iterable, attr, default=None):
    keyfunc = def_itemgetter(attr, default)
    data = list(iterable)
    order = unique_everseen(data, keyfunc)
    sorted_iterable = sorted(data, key=keyfunc)
    grouped = it.groupby(sorted_iterable, keyfunc)
    groups = {str(k): list(v) for k, v in grouped}

    # return groups in original order
    return ((key, groups[key]) for key in order)


def unique_everseen(iterable, key=None):
    # List unique elements, preserving order. Remember all elements ever seen
    # unique_everseen('ABBCcAD', str.lower) --> a b c d
    seen = set()

    for element in iterable:
        k = str(key(element))

        if k not in seen:
            seen.add(k)
            yield k


def betwix(iterable, start=None, stop=None, inc=False):
    """Extract selected elements from an iterable. But unlike `islice`,
    extract based on the element's value instead of its position.

    Args:
        iterable (iter): The initial sequence
        start (str): The fragment to begin with (inclusive)
        stop (str): The fragment to finish at (exclusive)
        inc (bool): Make stop operate inclusively (useful if reading a file and
            the start and stop fragments are on the same line)

    Returns:
        Iter: New dict with specified keys removed

    Examples:
        >>> from io import StringIO
        >>>
        >>> list(betwix('ABCDEFG', stop='C')) == ['A', 'B']
        True
        >>> list(betwix('ABCDEFG', 'C', 'E')) == ['C', 'D']
        True
        >>> list(betwix('ABCDEFG', 'C')) == ['C', 'D', 'E', 'F', 'G']
        True
        >>> f = StringIO('alpha\\n<beta>\\ngamma\\n')
        >>> list(betwix(f, '<', '>', True)) == ['<beta>\\n']
        True
        >>> list(betwix('ABCDEFG', 'C', 'E', True)) == ['C', 'D', 'E']
        True
    """

    def inc_takewhile(predicate, _iter):
        for x in _iter:
            yield x

            if not predicate(x):
                break

    get_pred = lambda sentinel: lambda x: sentinel not in x
    pred = get_pred(stop)
    first = it.dropwhile(get_pred(start), iterable) if start else iterable

    if stop and inc:
        last = inc_takewhile(pred, first)
    elif stop:
        last = it.takewhile(pred, first)
    else:
        last = first

    return last


def dispatch(split, *funcs):
    """takes a tuple of items and delivers each item to a different function

           /--> item1 --> double(item1) -----> \
          /                                     \
    split ----> item2 --> triple(item2) -----> _OUTPUT
          \\                                     /
           \\--> item3 --> quadruple(item3) --> /

    One way to construct such a flow in code would be::

        split = ('bar', 'baz', 'qux')
        double = lambda word: word * 2
        triple = lambda word: word * 3
        quadruple = lambda word: word * 4
        _OUTPUT = dispatch(split, double, triple, quadruple)
        _OUTPUT == ('barbar', 'bazbazbaz', 'quxquxquxqux')
    """
    return [func(item) for item, func in zip(split, funcs)]


def broadcast(item, *funcs):
    """delivers the same item to different functions

           /--> item --> double(item) -----> \
          /                                   \
    item -----> item --> triple(item) -----> _OUTPUT
          \\                                   /
           \\--> item --> quadruple(item) --> /

    One way to construct such a flow in code would be::

        double = lambda word: word * 2
        triple = lambda word: word * 3
        quadruple = lambda word: word * 4
        _OUTPUT = broadcast('bar', double, triple, quadruple)
        _OUTPUT == ('barbar', 'bazbazbaz', 'quxquxquxqux')
    """
    return [func(item) for func in funcs]


def _gen_words(match, splits):
    groups = list(it.dropwhile(lambda x: not x, match.groups()))

    for s in splits:
        try:
            num = int(s)
        except ValueError:
            word = s
        else:
            word = next(it.islice(groups, num, num + 1))

        yield word


def multi_substitute(word, rules):
    """Apply multiple regex rules to 'word'
    http://code.activestate.com/recipes/
    576710-multi-regex-single-pass-replace-of-multiple-regexe/
    """
    flags = rules[0]["flags"]

    # Create a combined regex from the rules
    tuples = ((p, r["match"]) for p, r in enumerate(rules))
    regexes = ("(?P<match_%i>%s)" % (p, r) for p, r in tuples)
    pattern = "|".join(regexes)
    regex = re.compile(pattern, flags)
    resplit = re.compile("\\$(\\d+)")

    # For each match, look-up corresponding replace value in dictionary
    rules_in_series = filter(itemgetter("series"), rules)
    rules_in_parallel = (r for r in rules if not r["series"])

    try:
        has_parallel = [next(rules_in_parallel)]
    except StopIteration:
        has_parallel = []

    # print('================')
    # pprint(rules)
    # print('word:', word)
    # print('pattern', pattern)
    # print('flags', flags)

    for _ in it.chain(rules_in_series, has_parallel):
        # print('~~~~~~~~~~~~~~~~')
        # print('new round')
        # print('word:', word)
        # found = list(regex.finditer(word))
        # matchitems = [match.groupdict().items() for match in found]
        # pprint(matchitems)
        prev_name = None
        prev_is_series = None
        i = 0

        for match in regex.finditer(word):
            items = match.groupdict().items()
            item = next(filter(itemgetter(1), items))

            # print('----------------')
            # print('groupdict:', match.groupdict().items())
            # print('item:', item)

            if not item:
                continue

            name = item[0]
            rule = rules[int(name[6:])]
            series = rule.get("series")
            kwargs = {"count": rule["count"], "series": series}
            is_previous = name == prev_name
            singlematch = kwargs["count"] == 1
            is_series = prev_is_series or kwargs["series"]
            isnt_previous = bool(prev_name) and not is_previous

            if (is_previous and singlematch) or (isnt_previous and is_series):
                continue

            prev_name = name
            prev_is_series = series

            if resplit.findall(rule["replace"]):
                splits = resplit.split(rule["replace"])
                words = _gen_words(match, splits)
            else:
                splits = rule["replace"]
                start = match.start() + i
                end = match.end() + i
                words = [word[:start], splits, word[end:]]
                i += rule["offset"]

            word = "".join(words)

            # print('name:', name)
            # print('prereplace:', rule['replace'])
            # print('splits:', splits)
            # print('resplits:', resplit.findall(rule['replace']))
            # print('groups:', filter(None, match.groups()))
            # print('i:', i)
            # print('words:', words)
            # print('range:', match.start(), '-', match.end())
            # print('replace:', word)

    # print('substitution:', word)
    return word


def substitute(word, rule):
    if word:
        result = rule["match"].subn(rule["replace"], word, rule["count"])
        replaced, replacements = result

        if rule.get("default") is not None and not replacements:
            replaced = rule.get("default")
    else:
        replaced = word

    return replaced


def get_new_rule(rule, recompile=False):
    flags = 0 if rule.get("casematch") else re.IGNORECASE

    if not rule.get("singlelinematch"):
        flags |= re.MULTILINE
        flags |= re.DOTALL

    count = 1 if rule.get("singlematch") else 0

    if recompile and "$" in rule["replace"]:
        replace = re.sub(r"\$(\d+)", r"\\\1", rule["replace"], 0)
    else:
        replace = rule["replace"]

    match = re.compile(rule["match"], flags) if recompile else rule["match"]

    nrule = {
        "match": match,
        "replace": replace,
        "default": rule.get("default"),
        "field": rule.get("field"),
        "count": count,
        "flags": flags,
        "series": rule.get("seriesmatch", True),
        "offset": int(rule.get("offset") or 0),
    }

    return nrule


def multiplex(sources):
    """Combine multiple generators into one"""
    return it.chain.from_iterable(sources)


def gen_entries(parsed):
    if parsed.get("bozo_exception"):
        raise Exception(parsed["bozo_exception"])

    for entry in parsed["entries"]:
        # prevent feedparser deprecation warnings
        if "published_parsed" in entry:
            updated = entry["published_parsed"]
        else:
            updated = entry.get("updated_parsed")

        entry["pubDate"] = updated
        entry["y:published"] = updated
        entry["dc:creator"] = entry.get("author")
        entry["author.uri"] = entry.get("author_detail", {}).get("href")
        entry["author.name"] = entry.get("author_detail", {}).get("name")
        entry["y:title"] = entry.get("title")
        entry["y:id"] = entry.get("id")
        yield entry


def gen_items(content, key=None):
    if hasattr(content, "append"):
        for nested in content:
            for i in gen_items(nested, key):
                yield i
    elif content:
        yield {key: content} if key else content



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.timeout
~~~~~~~~~~~~~~~~~~~~
Provides functions for returning items from a stream until a certain amount of
time has passed.

Contrast this with the truncate module, which also limits the number of items,
but returns items based on a count.

Examples:
    basic usage::

        >>> from time import sleep
        >>> from riko.modules.timeout import pipe
        >>>
        >>> def gen_items():
        ...     for x in range(50):
        ...         sleep(1)
        ...         yield {'x': x}
        >>>
        >>> len(list(pipe(gen_items(), conf={'seconds': '3'})))
        3

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
import signal

from datetime import timedelta

from . import operator
import pygogo as gogo

OPTS = {"ptype": "int"}
DEFAULTS = {}
logger = gogo.Gogo(__name__, monolog=True).logger


class TimeoutIterator(object):
    def __init__(self, elements, timeout=0):
        self.iter = iter(elements)
        self.timeout = timeout
        self.timedout = False
        self.started = False

    def _handler(self, signum, frame):
        self.timedout = True

    def __iter__(self):
        return self

    def __next__(self):
        if self.timedout:
            raise StopIteration
        elif not self.started:
            signal.signal(signal.SIGALRM, self._handler)
            signal.alarm(self.timeout)
            self.started = True

        try:
            return next(self.iter)
        except StopIteration:
            signal.alarm(0)
            self.timedout = True
            raise StopIteration


def parser(stream, objconf, tuples, **kwargs):
    """Parses the pipe content

    Args:
        stream (Iter[dict]): The source. Note: this shares the `tuples`
            iterator, so consuming it will consume `tuples` as well.

        objconf (obj): the item independent configuration (an Objectify
            instance).

        tuples (Iter[(dict, obj)]): Iterable of tuples of (item, objconf)
            `item` is an element in the source stream and `objconf` is the item
            configuration (an Objectify instance). Note: this shares the
            `stream` iterator, so consuming it will consume `stream` as well.

        kwargs (dict): Keyword arguments.

    Returns:
        Iter(dict): The output stream

    Examples:
        >>> from time import sleep
        >>> from meza.fntools import Objectify
        >>> from itertools import repeat, count
        >>>
        >>> kwargs = {'seconds': 3}
        >>> objconf = Objectify(kwargs)
        >>>
        >>> def gen_stream():
        ...     for x in count():
        ...         sleep(1)
        ...         yield {'x': x}
        >>>
        >>> stream = gen_stream()
        >>> tuples = zip(stream, repeat(objconf))
        >>> len(list(parser(stream, objconf, tuples, **kwargs)))
        3
    """
    time = int(timedelta(**objconf).total_seconds())
    return TimeoutIterator(stream, time)


@operator(DEFAULTS, isasync=True, **OPTS)
def async_pipe(*args, **kwargs):
    """An operator that asynchronously returns items from a stream until a
        certain amount of time has passed.

    Args:
        items (Iter[dict]): The source.
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain any of the following
            keys: 'days', 'seconds', 'microseconds', 'milliseconds', 'minutes',
            'hours', 'weeks'.

            days (int): Number of days before signaling a timeout (default: 0)
            seconds (int): Number of seconds before signaling a timeout
                (default: 0)
            microseconds (int): Number of microseconds before signaling a
                timeout (default: 0)
            milliseconds (int): Number of milliseconds before signaling a
                timeout (default: 0)
            minutes (int): Number of minutes before signaling a timeout
                (default: 0)
            hours (int): Number of hours before signaling a timeout
                (default: 0)
            weeks (int): Number of weeks before signaling a timeout
                (default: 0)

    Returns:
        Deferred: twisted.internet.defer.Deferred stream

    Examples:
        >>> from time import sleep
        >>> from riko.bado import react
        >>> from riko.bado.mock import FakeReactor
        >>>
        >>> def gen_items():
        ...     for x in range(50):
        ...         sleep(1)
        ...         yield {'x': x}
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(len(list(x)))
        ...     d = async_pipe(gen_items(), conf={'seconds': '3'})
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        3
    """
    return parser(*args, **kwargs)


@operator(DEFAULTS, **OPTS)
def pipe(*args, **kwargs):
    """An operator that returns items from a stream until a certain amount of
        time has passed.

    Args:
        items (Iter[dict]): The source.
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain any of the following
            keys: 'days', 'seconds', 'microseconds', 'milliseconds', 'minutes',
            'hours', 'weeks'.

            days (int): Number of days before signaling a timeout (default: 0)
            seconds (int): Number of seconds before signaling a timeout
                (default: 0)
            microseconds (int): Number of microseconds before signaling a
                timeout (default: 0)
            milliseconds (int): Number of milliseconds before signaling a
                timeout (default: 0)
            minutes (int): Number of minutes before signaling a timeout
                (default: 0)
            hours (int): Number of hours before signaling a timeout
                (default: 0)
            weeks (int): Number of weeks before signaling a timeout
                (default: 0)

    Yields:
        dict: an item

    Examples:
        >>> from time import sleep
        >>>
        >>> def gen_items():
        ...     for x in range(50):
        ...         sleep(1)
        ...         yield {'x': x}
        >>>
        >>> len(list(pipe(gen_items(), conf={'seconds': '3'})))
        3
    """
    return parser(*args, **kwargs)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.cast
~~~~~~~~~
Provides type casting capabilities
"""
from decimal import Decimal, InvalidOperation
from json import loads
from operator import add, sub
from time import gmtime
from datetime import timedelta
from calendar import timegm
from urllib.parse import quote, urlparse
from ast import literal_eval

from dateutil import parser
from meza.compat import decode
from riko.dates import TODAY, gen_tzinfos, get_date, normalize_date, get_tt
from riko.currencies import CURRENCY_CODES
from riko.locations import LOCATIONS

URL_SAFE = "%/:=&?~#+!$,;'@()*[]"
MATH_WORDS = {"seconds", "minutes", "hours", "days", "weeks", "months", "years"}
TEXT_WORDS = {"last", "next", "week", "month", "year"}
TT_KEYS = (
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "second",
    "day_of_week",
    "day_of_year",
    "daylight_savings",
)

DATES = {
    "today": TODAY,
    "now": TODAY,
    "tomorrow": TODAY + timedelta(days=1),
    "yesterday": TODAY - timedelta(days=1),
}

TZINFOS = dict(gen_tzinfos())

url_quote = lambda url: quote(url, safe=URL_SAFE)


def literal_parse(string):
    if string.lower() in {"true", "false"}:
        parsed = loads(string.lower())
    else:
        try:
            parsed = literal_eval(string)
        except (ValueError, SyntaxError):
            parsed = string

    return parsed


def cast_url(url_str):
    url = "http://%s" % url_str if "://" not in url_str else url_str
    quoted = url_quote(url)
    parsed = urlparse(quoted)
    response = parsed._asdict()
    response["url"] = parsed.geturl()
    return response


def lookup_street_address(address):
    location = {
        "lat": 0,
        "lon": 0,
        "country": "United States",
        "admin1": "state",
        "admin2": "county",
        "admin3": "city",
        "city": "city",
        "street": "street",
        "postal": "61605",
    }

    return location


def lookup_ip_address(address):
    location = {
        "country": "United States",
        "admin1": "state",
        "admin2": "county",
        "admin3": "city",
        "city": "city",
    }

    return location


def lookup_coordinates(lat, lon):
    location = {
        "lat": lat,
        "lon": lon,
        "country": "United States",
        "admin1": "state",
        "admin2": "county",
        "admin3": "city",
        "city": "city",
        "street": "street",
        "postal": "61605",
    }

    return location


def cast_location(address, loc_type="street_address"):
    GEOLOCATERS = {
        "coordinates": lambda x: lookup_coordinates(*x),
        "street_address": lambda x: lookup_street_address(x),
        "ip_address": lambda x: lookup_ip_address(x),
        "currency": lambda x: CURRENCY_CODES.get(x, {}),
    }

    result = GEOLOCATERS[loc_type](address)

    if result.get("location"):
        extra = LOCATIONS.get(result["location"], {})
        result.update(extra)

    return result


def cast_date(date_str):
    try:
        words = date_str.split(" ")
    except AttributeError:
        date = gmtime(date_str) if hasattr(date_str, "real") else date_str
    else:
        mathish = set(words).intersection(MATH_WORDS)
        textish = set(words).intersection(TEXT_WORDS)

        if date_str[0] in {"+", "-"} and len(mathish) == 1:
            op = sub if date_str.startswith("-") else add
            date = get_date(mathish, words[0][1:], op)
        elif len(textish) == 2:
            date = get_date("%ss" % words[1], 1, add)
        elif date_str in DATES:
            date = DATES.get(date_str)
        else:
            date = parser.parse(date_str, tzinfos=TZINFOS)

    if date:
        normal = normalize_date(date)
        tt = get_tt(normal)

        # Make Sunday the first day of the week
        day_of_w = 0 if tt[6] == 6 else tt[6] + 1
        isdst = None if tt[8] == -1 else bool(tt[8])
        result = {"utime": timegm(tt), "timezone": "UTC", "date": normal}
        result.update(zip(TT_KEYS, tt))  # pylint: disable=W1637
        result.update({"day_of_week": day_of_w, "daylight_savings": isdst})
    else:
        result = {}

    return result


CAST_SWITCH = {
    "float": {"default": float("nan"), "func": float},
    "decimal": {"default": Decimal("NaN"), "func": Decimal},
    "int": {"default": 0, "func": lambda i: int(float(i))},
    "text": {"default": "", "func": decode},
    "date": {"default": {"date": TODAY}, "func": cast_date},
    "url": {"default": {}, "func": cast_url},
    "location": {"default": {}, "func": cast_location},
    "bool": {"default": False, "func": lambda i: bool(literal_parse(i))},
    "pass": {"default": None, "func": lambda i: i},
    "none": {"default": None, "func": lambda _: None},
}


def cast(content, _type="text", **kwargs):
    """Convert content from one type to another

    Args:
        content: The entry to convert

    Kwargs:
        _type (str): The type to convert to

    Returns:
        any: The converted content

    Examples:
        >>> content = '12.25'
        >>> cast(content, 'float')
        12.25
        >>> cast(content, 'decimal')
        Decimal('12.25')
        >>> cast(content, 'int')
        12
        >>> cast(content, 'text')
        '12.25'
        >>> cast(content, 'bool')
        True
        >>> cast('foo', 'float')
        nan
        >>> cast('foo', 'decimal')
        Decimal('NaN')
        >>> cast('foo', 'int')
        0
    """
    if content is None:
        value = CAST_SWITCH[_type]["default"]
    elif kwargs:
        value = CAST_SWITCH[_type]["func"](content, **kwargs)
    else:
        try:
            value = CAST_SWITCH[_type]["func"](content)
        except (InvalidOperation, ValueError):
            value = 0 if _type == "int" else CAST_SWITCH[_type]["func"]("NaN")

    return value



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
tests.test_examples
~~~~~~~~~~~~~~~~~~~

Provides example pipeline tests.
"""
import nose.tools as nt

from importlib import import_module
from decimal import Decimal


def setup_module():
    """site initialization"""
    global initialized
    initialized = True
    print("Basic Module Setup\n")


class TestExamples(object):
    def __init__(self):
        self.cls_initialized = False

    def _get_pipeline(self, pipe_name):
        module = import_module("examples.%s" % pipe_name)
        pipeline = module.pipe(test=True)
        return list(pipeline)

    def test_kazeeki(self):
        """Tests the kazeeki pipeline"""
        pipe_name = "kazeeki"
        pipeline = self._get_pipeline(pipe_name)
        example = {
            "description": (
                "We are looking for freelancers ( individuals and companies ) "
                "who offer their services related to Architecture Walkthrough "
                "and 3D animations. Please consider this job as a potential "
                "to several more and a long term relationship.   We are a "
                "Media"
            ),
            "id": "www.elance.com-66963214",
            "k:budget": Decimal("12.5"),
            "k:budget_full": "$12.50 / hr",
            "k:budget_w_sym": "$12.50",
            "k:categories": [
                {"content": "animation"},
                {"content": "design & multimedia"},
            ],
            "k:client_location": "Cambodia",
            "k:cur_code": "USD",
            "k:due": "Thu, 05 Feb 2015 11:46:40 EST",
            "k:job_type": "hourly",
            "k:num_jobs": "0",
            "k:per_awarded": "0%",
            "k:source": "www.elance.com",
            "k:posted": "Tue, 06 Jan 2015 11:46:40 EST",
            "k:submissions": "0",
            "k:tags": [
                {"content": "3d animation"},
                {"content": "3d modeling"},
                {"content": "3d rendering"},
                {"content": "animation"},
                {"content": "computer graphics"},
            ],
            "k:tot_purchased": "$0",
            "link": (
                "https://www.elance.com/j/3d-architecture-walkthrough-3d-"
                "animation-artists/66963214/"
            ),
            "title": ("3D Architecture Walkthrough &amp; 3D / Animation Artists"),
        }

        expected = 180
        length = len(pipeline)
        msg = "Pipeline %s has length %i, not %i"
        nt.assert_equal(length, expected, msg % (pipe_name, length, expected))
        nt.assert_equal(example, pipeline[-1])

    def test_gigs(self):
        """Tests the gigs pipeline"""
        pipe_name = "gigs"
        pipeline = self._get_pipeline(pipe_name)

        example = {
            "description": (
                "<b>Description:</b> Need a to port our existing iOS App to "
                "Android. Actually it is as good as writing a new Android "
                "app...<br><b>Category:</b> Web, Software & IT<br><b>Required "
                "skills:</b> android, sqlite<br><b>Fixed Price budget:</b> "
                "$250-$500<br><b>Project type:</b> Public<br><b>Freelancer "
                "Location:</b> India<br>"
            ),
            "guid": "http://www.guru.com/jobs/educational-android-app/1058980",
            "link": "http://www.guru.com/jobs/educational-android-app/1058980",
            "pubDate": "Tue, 05 Aug 2014 09:35:28 GMT",
            "title": "Educational Android App",
            "y:id": {
                "permalink": "false",
                "value": (
                    "http://www.guru.com/jobs/educational-android-app/" "1058980"
                ),
            },
            "y:published": {
                "day": "5",
                "day_name": "Tuesday",
                "day_of_week": "2",
                "day_ordinal_suffix": "th",
                "hour": "9",
                "minute": "35",
                "month": "8",
                "month_name": "August",
                "second": "28",
                "timezone": "UTC",
                "utime": "1407231328",
                "year": "2014",
            },
            "y:repeatcount": "1",
            "y:title": "Educational Android App",
        }

        length = len(pipeline)
        msg = "Pipeline %s has length %i, not 1"
        nt.assert_equal(length, 49, msg % (pipe_name, length))
        nt.assert_equal(example, pipeline[-1])

    def test_simple1(self):
        """Tests the simple1 pipeline"""
        pipe_name = "simple1"
        pipeline = self._get_pipeline(pipe_name)
        length = len(pipeline)
        msg = "Pipeline %s has length %i, not 1"
        nt.assert_equal(length, 1, msg % (pipe_name, length))
        nt.assert_equal({"url": "farechart"}, pipeline[-1])

    def test_simple2(self):
        """Tests the simple2 pipeline"""
        pipe_name = "simple2"
        pipeline = self._get_pipeline(pipe_name)
        example = {"author": "ABC", "link": "www.google.com", "title": "google"}

        length = len(pipeline)
        msg = "Pipeline %s has length %i, not 1"
        nt.assert_equal(length, 1, msg % (pipe_name, length))
        nt.assert_equal(example, pipeline[-1])

    def test_split(self):
        """Tests the split pipeline"""
        pipe_name = "split"
        pipeline = self._get_pipeline(pipe_name)
        example = {"date": "December 02, 2014", "year": 2014}
        length = len(pipeline)
        msg = "Pipeline %s has length %i, not 1"
        nt.assert_equal(length, 1, msg % (pipe_name, length))
        nt.assert_equal(example, pipeline[-1])

    def test_wired(self):
        """Tests the wired pipeline"""
        pipe_name = "wired"
        pipeline = self._get_pipeline(pipe_name)
        length = len(pipeline)
        msg = "Pipeline %s has length %i, not 1"
        nt.assert_equal(length, 1, msg % (pipe_name, length))
        nt.assert_equal({"date": "May 04, 1982"}, pipeline[-1])



from pprint import pprint
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

p1_conf = {
    "attrs": [{"value": "http://www.caltrain.com/Fares/farechart.html", "key": "url"}]
}

p2_conf = {"rule": {"field": "url", "match": {"subkey": "url"}, "replace": "farechart"}}


def pipe(test=False):
    stream = SyncPipe("itembuilder", conf=p1_conf, test=test).regex(conf=p2_conf).list

    for i in stream:
        pprint(str(i["url"]))

    return stream


@coroutine
def async_pipe(reactor, test=False):
    stream = yield (
        AsyncPipe("itembuilder", conf=p1_conf, test=test).regex(conf=p2_conf).list
    )

    for i in stream:
        pprint(str(i["url"]))



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.exchangerate
~~~~~~~~~~~~~~~~~~~~~~~~~
Provides functions for querying currency exchange rates

Examples:
    basic usage::

        >>> from riko import get_path
        >>> from riko.modules.exchangerate import pipe
        >>>
        >>> url = get_path('quote.json')
        >>> next(pipe({'content': 'GBP'}, conf={'url': url}))['exchangerate']
        Decimal('1.275201')

Attributes:
    OPTS (dict): The default pipe options
    DEFAULTS (dict): The default parser options
"""
import traceback

import pygogo as gogo

from json import loads
from decimal import Decimal
from os import getenv

from ijson import items
from meza.compat import decode

from . import processor
from riko.bado import coroutine, return_value, requests as treq, io
from riko.utils import fetch, get_abspath

EXCHANGE_API = "https://openexchangerates.org/api/latest.json"
PARAMS = {"app_id": getenv("OPEN_EXCHANGE_RATES_APP_ID")}

OPTS = {"field": "content", "ftype": "text"}
DEFAULTS = {
    "currency": "USD",
    "delay": 0,
    "memoize": True,
    "precision": 6,
    "url": EXCHANGE_API,
    "params": PARAMS,
}

logger = gogo.Gogo(__name__, monolog=True).logger


def parse_response(json):
    if "rates" in json:
        resp = {k: Decimal(v) for k, v in json["rates"].items() if v}
    else:
        # TODO: make sure this log shows up in console
        logger.warning("invalid json response:")
        logger.warning(json)
        resp = {}

    return resp


def calc_rate(from_cur, to_cur, rates, places=Decimal("0.0001")):
    def get_rate(currency):
        rate = rates.get(currency, Decimal("nan"))

        if not rate:
            logger.warning("rate USD/%s not found in rates" % currency)

        return rate

    if from_cur == to_cur:
        rate = Decimal(1)
    elif to_cur == "USD":
        rate = get_rate(from_cur)
    else:
        usd_to_given = get_rate(from_cur)
        usd_to_default = get_rate(to_cur)
        rate = usd_to_given / usd_to_default

    return (Decimal(1) / rate).quantize(places)


@coroutine
def async_parser(base, objconf, skip=False, **kwargs):
    """Asynchronously parses the pipe content

    Args:
        base (str): The base currency (exchanging from)
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: exchangerate)
        stream (dict): The original item

    Returns:
        Deferred: twisted.internet.defer.Deferred item

    Examples:
        >>> from riko import get_path
        >>> from riko.bado import react
        >>> from riko.bado.mock import FakeReactor
        >>> from meza.fntools import Objectify
        >>>
        >>> def run(reactor):
        ...     url = get_path('quote.json')
        ...     conf = {
        ...         'url': url, 'currency': 'USD', 'delay': 0, 'precision': 6}
        ...     item = {'content': 'GBP'}
        ...     objconf = Objectify(conf)
        ...     kwargs = {'stream': item, 'assign': 'content'}
        ...     d = async_parser(item['content'], objconf, **kwargs)
        ...     return d.addCallbacks(print, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        1.275201
    """
    same_currency = base == objconf.currency

    if skip:
        rate = kwargs["stream"]
    elif same_currency:
        rate = Decimal(1)
    elif objconf.url.startswith("http"):
        r = yield treq.get(objconf.url, params=objconf.params)
        json = yield treq.json(r)
    else:
        url = get_abspath(objconf.url)
        content = yield io.async_url_read(url, delay=objconf.delay)
        json = loads(decode(content))

    if not (skip or same_currency):
        places = Decimal(10) ** -objconf.precision
        rates = parse_response(json)
        rate = calc_rate(base, objconf.currency, rates, places=places)

    return_value(rate)


def parser(base, objconf, skip=False, **kwargs):
    """Parses the pipe content

    Args:
        base (str): The base currency (exchanging from)
        objconf (obj): The pipe configuration (an Objectify instance)
        skip (bool): Don't parse the content
        kwargs (dict): Keyword arguments

    Kwargs:
        assign (str): Attribute to assign parsed content (default: exchangerate)
        stream (dict): The original item

    Returns:
        dict: The item

    Examples:
        >>> from riko import get_path
        >>> from meza.fntools import Objectify
        >>>
        >>> url = get_path('quote.json')
        >>> conf = {'url': url, 'currency': 'USD', 'delay': 0, 'precision': 6}
        >>> item = {'content': 'GBP'}
        >>> objconf = Objectify(conf)
        >>> kwargs = {'stream': item, 'assign': 'content'}
        >>> parser(item['content'], objconf, **kwargs)
        Decimal('1.275201')
    """
    same_currency = base == objconf.currency

    if skip:
        rate = kwargs["stream"]
    elif same_currency:
        rate = Decimal(1)
    else:
        decode = objconf.url.startswith("http")

        with fetch(decode=decode, **objconf) as f:
            try:
                json = next(items(f, ""))
            except Exception as e:
                f.seek(0)
                logger.error("Error parsing {url}".format(**objconf))
                logger.debug(f.read())
                logger.error(e)
                logger.error(traceback.format_exc())
                skip = True
                rate = 0

    if not (skip or same_currency):
        places = Decimal(10) ** -objconf.precision
        rates = parse_response(json)
        rate = calc_rate(base, objconf.currency, rates, places=places)

    return rate


@processor(DEFAULTS, isasync=True, **OPTS)
def async_pipe(*args, **kwargs):
    """A processor that asynchronously retrieves the current exchange rate
    for a given currency pair.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain the keys 'url',
            'params', 'currency', 'delay', 'memoize', or 'field'.

            url (str): The exchange rate API url (default:
                http://finance.yahoo.com...)

            params (dict): The API url parameters (default: {'format': 'json'})
            currency: The (exchanging to) currency ISO abbreviation (default:
                USD).

            delay (flt): Amount of time to sleep (in secs) before fetching the
                url. Useful for simulating network latency. Default: 0.

            memoize (bool): Cache the exchange rate API response (default:
                True).

        field (str): Item attribute from which to obtain the string to be
            formatted (default: 'content')

        assign (str): Attribute to assign parsed content (default:
            exchangerate)

    Returns:
        dict: twisted.internet.defer.Deferred stream of items

    Examples:
        >>> from riko import get_path
        >>> from riko.bado import react
        >>> from riko.bado.mock import FakeReactor
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x)['exchangerate'])
        ...     url = get_path('quote.json')
        ...     d = async_pipe({'content': 'GBP'}, conf={'url': url})
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        1.275201
    """
    return async_parser(*args, **kwargs)


@processor(DEFAULTS, **OPTS)
def pipe(*args, **kwargs):
    """A processor that retrieves the current exchange rate for a given
    currency pair.

    Args:
        item (dict): The entry to process
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        conf (dict): The pipe configuration. May contain the keys 'url',
            'params', 'currency', 'delay', 'memoize', or 'field'.

            url (str): The exchange rate API url (default:
                http://finance.yahoo.com...)

            params (dict): The API url parameters (default: {'format': 'json'})
            currency: The (exchanging to) currency ISO abbreviation (default:
                USD).

            delay (flt): Amount of time to sleep (in secs) before fetching the
                url. Useful for simulating network latency. Default: 0.

            memoize (bool): Cache the exchange rate API response (default:
                True).

        field (str): Item attribute from which to obtain the string to be
            formatted (default: 'content')

        assign (str): Attribute to assign parsed content (default:
            exchangerate)

    Yields:
        dict: an item of the result

    Examples:
        >>> from riko import get_path
        >>> url = get_path('quote.json')
        >>> conf = {'url': url}
        >>> rate = next(pipe({'content': 'GBP'}, conf=conf))['exchangerate']
        >>> rate
        Decimal('1.275201')
        >>> msg = 'There are 1.28 GBPs per USD'
        >>> 'There are %#.2f GBPs per USD' % rate == msg
        True
        >>> conf = {'url': url, 'currency': 'TZS', 'precision': 3}
        >>> next(pipe({'content': 'USD'}, conf=conf))['exchangerate']
        Decimal('2282.466')
        >>> conf = {'url': url, 'currency': 'XYZ'}
        >>> next(pipe({'content': 'USD'}, conf=conf))['exchangerate']
        Decimal('NaN')
    """
    return parser(*args, **kwargs)



from pprint import pprint
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

p232_conf = {
    "attrs": [
        {"value": "www.google.com", "key": "link"},
        {"value": "google", "key": "title"},
        {"value": "empty", "key": "author"},
    ]
}

p421_conf = {"rule": [{"find": "empty", "param": "first", "replace": "ABC"}]}


def pipe(test=False):
    stream = (
        SyncPipe("itembuilder", conf=p232_conf, test=test)
        .strreplace(conf=p421_conf, field="author", assign="author")
        .list
    )

    for i in stream:
        pprint(i)

    return stream


@coroutine
def async_pipe(reactor, test=False):
    stream = yield (
        AsyncPipe("itembuilder", conf=p232_conf, test=test)
        .strreplace(conf=p421_conf, field="author", assign="author")
        .list
    )

    for i in stream:
        pprint(i)



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
riko.modules.udf
~~~~~~~~~~~~~~~~
Provides functions for performing an arbitrary (user-defined) function on stream
items.

Examples:
    basic usage::

        >>> from riko.modules.udf import pipe
        >>>
        >>> items = [{'x': x} for x in range(5)]
        >>> func = lambda item: {'y': item['x'] + 3}
        >>> next(pipe(items, func=func))
        {'y': 3}
"""
from . import operator
import pygogo as gogo

logger = gogo.Gogo(__name__, monolog=True).logger


def parser(stream, objconf, tuples, **kwargs):
    """Parses the pipe content

    Args:
        stream (Iter[dict]): The source. Note: this shares the `tuples`
            iterator, so consuming it will consume `tuples` as well.

        objconf (obj): the item independent configuration (an Objectify
            instance).

        tuples (Iter[(dict, obj)]): Iterable of tuples of (item, objconf)
            `item` is an element in the source stream and `objconf` is the item
            configuration (an Objectify instance). Note: this shares the
            `stream` iterator, so consuming it will consume `stream` as well.

        kwargs (dict): Keyword arguments.

    Returns:
        Iter(dict): The output stream

    Examples:
        >>> from meza.fntools import Objectify
        >>> from itertools import repeat
        >>>
        >>> func = lambda item: {'y': item['x'] + 3}
        >>> stream = ({'x': x} for x in range(5))
        >>> tuples = zip(stream, repeat(None))
        >>> next(parser(stream, None, tuples, func=func))
        {'y': 3}
    """
    return map(kwargs["func"], stream)


@operator(isasync=True)
def async_pipe(*args, **kwargs):
    """An operator that asynchronously performs an arbitrary (user-defined) function on
    items of a stream.

    Args:
        items (Iter[dict]): The source.
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        func (callable): User defined function to apply to each stream item.

    Returns:
        Deferred: twisted.internet.defer.Deferred truncated stream

    Examples:
        >>> from riko.bado import react
        >>> from riko.bado.mock import FakeReactor
        >>>
        >>> def run(reactor):
        ...     callback = lambda x: print(next(x))
        ...     func = lambda item: {'y': item['x'] + 3}
        ...     items = ({'x': x} for x in range(5))
        ...     d = async_pipe(items, func=func)
        ...     return d.addCallbacks(callback, logger.error)
        >>>
        >>> try:
        ...     react(run, _reactor=FakeReactor())
        ... except SystemExit:
        ...     pass
        ...
        {'y': 3}
    """
    return parser(*args, **kwargs)


@operator()
def pipe(*args, **kwargs):
    """An operator that performs an arbitrary (user-defined) function on items of a
    stream.

    Args:
        items (Iter[dict]): The source.
        kwargs (dict): The keyword arguments passed to the wrapper

    Kwargs:
        func (callable): User defined function to apply to each stream item.

    Yields:
        dict: an item

    Examples:
        >>> items = [{'x': x} for x in range(5)]
        >>> func = lambda item: {'y': item['x'] + 3}
        >>> next(pipe(items, func=func))
        {'y': 3}
    """
    return parser(*args, **kwargs)



from pprint import pprint
from riko import get_path
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

p1_conf = {"url": get_path("gigs.json"), "path": "value.items"}
p2_conf = {"uniq_key": "link"}
p3_conf = {
    "combine": "or",
    "mode": "block",
    "rule": [{"field": "title", "value": "php", "op": "contains"}],
}

p4_conf = {"rule": [{"sort_key": "pubDate", "sort_dir": "desc"}]}


def pipe(test=False):
    stream = (
        SyncPipe("fetchdata", conf=p1_conf, test=test)
        .uniq(conf=p2_conf)
        .filter(conf=p3_conf)
        .sort(conf=p4_conf)
        .list
    )

    for i in stream:
        pprint(i)

    return stream


@coroutine
def async_pipe(reactor, test=False):
    stream = yield (
        AsyncPipe("fetchdata", conf=p1_conf, test=test)
        .uniq(conf=p2_conf)
        .filter(conf=p3_conf)
        .sort(conf=p4_conf)
        .output
    )

    for i in stream:
        pprint(i)



from pprint import pprint
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

p385_conf = {"type": "date"}
p385_in = {"content": "12/2/2014"}
p405_conf = {"format": "%B %d, %Y"}
p393_conf = {
    "attrs": [
        {"value": {"terminal": "date", "path": "dateformat"}, "key": "date"},
        {"value": {"terminal": "year", "path": "year"}, "key": "year"},
    ]
}

p385_kwargs = {"conf": p385_conf, "inputs": p385_in}


def pipe(test=False):
    s1, s2 = (
        SyncPipe("input", test=test, **p385_kwargs)
        .dateformat(conf=p405_conf)
        .split()
        .output
    )

    p393_kwargs = {"conf": p393_conf, "date": s1, "year": s2, "test": test}
    stream = SyncPipe("itembuilder", **p393_kwargs).list

    for i in stream:
        pprint(i)

    return stream


@coroutine
def async_pipe(reactor, test=False):
    s1, s2 = yield (
        AsyncPipe("input", test=test, **p385_kwargs)
        .dateformat(conf=p405_conf)
        .split()
        .output
    )

    p393_kwargs = {"conf": p393_conf, "date": s1, "year": s2, "test": test}
    stream = yield AsyncPipe("itembuilder", **p393_kwargs).list

    for i in stream:
        pprint(i)



# -*- test-case-name: twisted.web.test.test_xml -*-
#
# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.


"""
*S*mall, *U*ncomplicated *X*ML.

This is a very simple implementation of XML/HTML as a network
protocol.  It is not at all clever.  Its main features are that it
does not:

  - support namespaces
  - mung mnemonic entity references
  - validate
  - perform *any* external actions (such as fetching URLs or writing files)
    under *any* circumstances
  - has lots and lots of horrible hacks for supporting broken HTML (as an
    option, they're not on by default).
"""

import pygogo as gogo

from chardet import detect
from meza.compat import decode

try:
    from twisted.internet.protocol import Protocol
except ImportError:
    Protocol = object
else:
    from twisted.python.reflect import prefixedMethodNames as find_method_names

logger = gogo.Gogo(__name__, monolog=True).logger

# Elements of the three-tuples in the state table.
BEGIN_HANDLER = 0
DO_HANDLER = 1
END_HANDLER = 2
IDENTCHARS = ".-_:"
LENIENT_IDENTCHARS = IDENTCHARS + ";+#/%~"

nop = lambda *args, **kwargs: None


def zipfndict(*args):
    for fndict in args:
        for key in fndict:
            yield (key, tuple(x.get(key, nop) for x in args))


def get_method_obj_dict(obj, prefix):
    names = find_method_names(obj.__class__, prefix)
    return {name: getattr(obj, prefix + name) for name in names}


class ParseError(Exception):
    def __init__(self, filename, line, col, message):
        self.filename = filename
        self.line = line
        self.col = col
        self.message = message

    def __str__(self):
        return "%s:%s:%s: %s" % (self.filename, self.line, self.col, self.message)


class XMLParser(Protocol):
    state = None
    encoding = None
    bom = None
    attrname = ""
    attrval = ""

    # _leadingBodyData will sometimes be set before switching to the
    # 'bodydata' state, when we "accidentally" read a byte of bodydata
    # in a different state.
    _leadingBodyData = None

    def __init__(self, filename="unnamed", **kwargs):
        self.filename = filename
        self.lenient = kwargs.get("lenient")
        self.strict = not self.lenient

    # protocol methods
    def connectionMade(self):
        self.lineno = 1
        self.colno = 0

    def saveMark(self):
        """Get the line number and column of the last character parsed"""
        # This gets replaced during dataReceived, restored afterwards
        return (self.lineno, self.colno)

    def _raise_parse_error(self, message):
        raise ParseError(*((self.filename,) + self.saveMark() + (message,)))

    def _build_state_table(self):
        """Return a dictionary of begin, do, end state function tuples"""
        # _build_state_table leaves something to be desired but it does what it
        # does.. probably slowly, so I'm doing some evil caching so it doesn't
        # get called more than once per class.
        stateTable = getattr(self.__class__, "__stateTable", None)

        if stateTable is None:
            prefixes = ("begin_", "do_", "end_")
            fndicts = (get_method_obj_dict(self, p) for p in prefixes)
            stateTable = dict(zipfndict(*fndicts))
            self.__class__.__stateTable = stateTable

        return stateTable

    def check_encoding(self, data):
        if self.encoding.startswith("UTF-16"):
            data = data[2:]

        if "UTF-16" in self.encoding or "UCS-2" in self.encoding:
            assert not len(data) & 1, "UTF-16 must come in pairs for now"

    def maybeBodyData(self):
        if self.endtag:
            return "bodydata"

        # Get ready for fun! We're going to allow
        # <script>if (foo < bar)</script> to work!
        # We do this by making everything between <script> and
        # </script> a Text
        # BUT <script src="foo"> will be special-cased to do regular,
        # lenient behavior, because those may not have </script>
        # -radix
        if self.tagName == "script" and "src" not in self.tagAttributes:
            # we do this ourselves rather than having begin_waitforendscript
            # because that can get called multiple times and we don't want
            # bodydata to get reset other than the first time.
            self.begin_bodydata(None)
            return "waitforendscript"

        return "bodydata"

    def dataReceived(self, data):
        stateTable = self._build_state_table()
        self.encoding = self.encoding or detect(data)["encoding"]
        self.check_encoding(data)
        self.state = self.state or "begin"
        content = decode(data, self.encoding)

        # bring state, lineno, colno into local scope
        lineno, colno = self.lineno, self.colno
        curState = self.state

        # replace saveMark with a nested scope function
        saveMark = lambda: (lineno, colno)
        self.saveMark, _saveMark = saveMark, self.saveMark

        # fetch functions from the stateTable
        beginFn, doFn, endFn = stateTable[curState]

        try:
            for char in content:
                # do newline stuff
                if char == "\n":
                    lineno += 1
                    colno = 0
                else:
                    colno += 1

                newState = doFn(char)

                if newState and newState != curState:
                    # this is the endFn from the previous state
                    endFn()
                    curState = newState
                    beginFn, doFn, endFn = stateTable[curState]
                    beginFn(char)
        finally:
            self.saveMark = _saveMark
            self.lineno, self.colno = lineno, colno

        # state doesn't make sense if there's an exception..
        self.state = curState

    def connectionLost(self, reason):
        """
        End the last state we were in.
        """
        stateTable = self._build_state_table()
        stateTable[self.state][END_HANDLER]()

    # state methods

    def do_begin(self, byte):
        if byte.isspace():
            return

        if byte != "<" and self.lenient:
            self._leadingBodyData = byte
            return "bodydata"
        elif byte != "<":
            msg = "First char of document [%r] wasn't <" % (byte,)
            self._raise_parse_error(msg)

        return "tagstart"

    def begin_comment(self, byte):
        self.commentbuf = ""

    def do_comment(self, byte):
        self.commentbuf += byte

        if self.commentbuf.endswith("-->"):
            self.gotComment(self.commentbuf[:-3])
            return "bodydata"

    def begin_tagstart(self, byte):
        self.tagName = ""  # name of the tag
        self.tagAttributes = {}  # attributes of the tag
        self.termtag = 0  # is the tag self-terminating
        self.endtag = 0

    def _get_val(self, byte):
        val = None
        alnum_or_ident = byte.isalnum() or byte in IDENTCHARS
        is_good = alnum_or_ident or byte in "/!?[" or byte.isspace()

        if byte == "-" and self.tagName == "!-":
            val = "comment"
        elif byte.isspace() and self.tagName:
            # properly strict thing to do here is probably to only
            # accept whitespace
            val = "waitforgt" if self.endtag else "attrs"
        elif byte in ">/[":
            def_gt = self.strict and "bodydata" or self.maybeBodyData()

            switch = {
                ">": "bodydata" if self.endtag else def_gt,
                "/": "afterslash" if self.tagName else None,
                "[": "expectcdata" if self.tagName == "!" else None,
            }

            val = switch[byte]

        if not (self.lenient or val or is_good):
            self._raise_parse_error("Invalid tag character: %r" % byte)

        return val

    def _update_tags(self, byte):
        alnum_or_ident = byte.isalnum() or byte in IDENTCHARS

        if (byte in "!?") or alnum_or_ident:
            self.tagName += byte
        elif byte == ">" and self.endtag:
            self.gotTagEnd(self.tagName)
        elif byte == ">":
            self.gotTagStart(self.tagName, {})
        elif byte == "/" and not self.tagName:
            self.endtag = 1
        elif byte in "!?" and not self.tagName:
            self.tagName += byte
            self.termtag = 1

    def do_tagstart(self, byte):
        if byte.isspace() and not self.tagName:
            self._raise_parse_error("Whitespace before tag-name")
        elif byte in "!?" and self.tagName and self.strict:
            self._raise_parse_error("Invalid character in tag-name")
        elif byte == "[" and not self.tagName == "!":
            self._raise_parse_error("Invalid '[' in tag-name")

        val = self._get_val(byte)
        self._update_tags(byte)
        return val

    def begin_unentity(self, byte):
        self.bodydata += byte

    def do_unentity(self, byte):
        self.bodydata += byte
        return "bodydata"

    def end_unentity(self):
        self.gotText(self.bodydata)

    def begin_expectcdata(self, byte):
        self.cdatabuf = byte

    def do_expectcdata(self, byte):
        self.cdatabuf += byte
        cdb = self.cdatabuf
        cd = "[CDATA["

        if len(cd) > len(cdb):
            if cd.startswith(cdb):
                return
            elif self.lenient:
                # WHAT THE CRAP!?  MSWord9 generates HTML that includes these
                # bizarre <![if !foo]> <![endif]> chunks, so I've gotta ignore
                # 'em as best I can.  this should really be a separate parse
                # state but I don't even have any idea what these _are_.
                return "waitforgt"
            else:
                self._raise_parse_error("Mal-formed CDATA header")
        if cd == cdb:
            self.cdatabuf = ""
            return "cdata"

        self._raise_parse_error("Mal-formed CDATA header")

    def do_cdata(self, byte):
        self.cdatabuf += byte
        if self.cdatabuf.endswith("]]>"):
            self.cdatabuf = self.cdatabuf[:-3]
            return "bodydata"

    def end_cdata(self):
        self.gotCData(self.cdatabuf)
        self.cdatabuf = ""

    def do_attrs(self, byte):
        if byte.isalnum() or byte in IDENTCHARS:
            # XXX FIXME really handle !DOCTYPE at some point
            if self.tagName == "!DOCTYPE":
                return "doctype"

            if self.tagName[0] in "!?":
                return "waitforgt"

            return "attrname"
        elif byte.isspace():
            return
        elif byte == ">":
            self.gotTagStart(self.tagName, self.tagAttributes)
            return self.strict and "bodydata" or self.maybeBodyData()
        elif byte == "/":
            return "afterslash"
        elif self.lenient:
            # discard and move on?  Only case I've seen of this so far was:
            # <foo bar="baz"">
            return

        self._raise_parse_error("Unexpected character: %r" % byte)

    def begin_doctype(self, byte):
        self.doctype = byte

    def do_doctype(self, byte):
        if byte == ">":
            return "bodydata"

        self.doctype += byte

    def end_doctype(self):
        self.gotDoctype(self.doctype)
        self.doctype = None

    def do_waitforgt(self, byte):
        if byte == ">":
            if self.endtag or self.lenient:
                return "bodydata"

            return self.maybeBodyData()

    def begin_attrname(self, byte):
        self.attrname = byte
        self._attrname_termtag = 0

    def _get_attrname(self, byte):
        if byte == "=":
            val = "beforeattrval"
        elif byte.isspace():
            val = "beforeeq"
        elif self.lenient and byte in "\"'":
            val = "attrval"
        elif self.lenient and byte == ">":
            val = "bodydata" if self._attrname_termtag else None
        else:
            # something is really broken. let's leave this attribute where it
            # is and move on to the next thing
            val = None

        return val

    def do_attrname(self, byte):
        if byte.isalnum() or byte in IDENTCHARS:
            self.attrname += byte
        elif self.strict and not (byte.isspace() or byte == "="):
            msg = "Invalid attribute name: %r %r" % (self.attrname, byte)
            self._raise_parse_error(msg)
        elif byte in LENIENT_IDENTCHARS or byte.isalnum():
            self.attrname += byte
        elif byte == "/":
            self._attrname_termtag = 1
        elif byte == ">":
            self.attrval = "True"
            self.tagAttributes[self.attrname] = self.attrval
            self.gotTagStart(self.tagName, self.tagAttributes)
            self.gotTagEnd(self.tagName) if self._attrname_termtag else None

        return self._get_attrname(byte)

    def do_beforeattrval(self, byte):
        chars = LENIENT_IDENTCHARS
        val = None

        if byte in "\"'":
            val = "attrval"
        elif byte.isspace():
            pass
        elif self.lenient and (byte in chars or byte.isalnum()):
            val = "messyattr"
        elif self.lenient and byte == ">":
            self.attrval = "True"
            self.tagAttributes[self.attrname] = self.attrval
            self.gotTagStart(self.tagName, self.tagAttributes)
            val = self.maybeBodyData()
        elif self.lenient and byte == "\\":
            # I saw this in actual HTML once:
            # <font size=\"3\"><sup>SM</sup></font>
            pass
        else:
            msg = "Invalid initial attribute value: %r; " % byte
            msg += "Attribute values must be quoted."
            self._raise_parse_error(msg)

        return val

    def begin_beforeeq(self, byte):
        self._beforeeq_termtag = 0

    def do_beforeeq(self, byte):
        if byte == "=":
            return "beforeattrval"
        elif byte.isspace():
            return
        elif self.lenient:
            if byte.isalnum() or byte in IDENTCHARS:
                self.attrval = "True"
                self.tagAttributes[self.attrname] = self.attrval
                return "attrname"
            elif byte == ">":
                self.attrval = "True"
                self.tagAttributes[self.attrname] = self.attrval
                self.gotTagStart(self.tagName, self.tagAttributes)

                if self._beforeeq_termtag:
                    self.gotTagEnd(self.tagName)
                    return "bodydata"

                return self.maybeBodyData()
            elif byte == "/":
                self._beforeeq_termtag = 1
                return

        self._raise_parse_error("Invalid attribute")

    def begin_attrval(self, byte):
        self.quotetype = byte
        self.attrval = ""

    def do_attrval(self, byte):
        if byte == self.quotetype:
            return "attrs"
        self.attrval += byte

    def end_attrval(self):
        self.tagAttributes[self.attrname] = self.attrval
        self.attrname = self.attrval = ""

    def begin_messyattr(self, byte):
        self.attrval = byte

    def do_messyattr(self, byte):
        if byte.isspace():
            return "attrs"
        elif byte == ">":
            endTag = 0

            if self.attrval.endswith("/"):
                endTag = 1
                self.attrval = self.attrval[:-1]

            self.tagAttributes[self.attrname] = self.attrval
            self.gotTagStart(self.tagName, self.tagAttributes)

            if endTag:
                self.gotTagEnd(self.tagName)
                return "bodydata"

            return self.maybeBodyData()
        else:
            self.attrval += byte

    def end_messyattr(self):
        if self.attrval:
            self.tagAttributes[self.attrname] = self.attrval

    def begin_afterslash(self, byte):
        self._after_slash_closed = 0

    def do_afterslash(self, byte):
        # this state is only after a self-terminating slash, e.g. <foo/>
        if self._after_slash_closed:
            self._raise_parse_error("Mal-formed")  # XXX When does this happen??

        if byte != ">" and self.lenient:
            return
        elif byte != ">":
            self._raise_parse_error("No data allowed after '/'")

        self._after_slash_closed = 1
        self.gotTagStart(self.tagName, self.tagAttributes)
        self.gotTagEnd(self.tagName)

        # don't need maybeBodyData here because there better not be
        # any javascript code after a <script/>... we'll see :(
        return "bodydata"

    def begin_bodydata(self, byte):
        if self._leadingBodyData:
            self.bodydata = self._leadingBodyData
            del self._leadingBodyData
        else:
            self.bodydata = ""

    def do_bodydata(self, byte):
        if byte == "<":
            return "tagstart"
        if byte == "&":
            return "entityref"
        self.bodydata += byte

    def end_bodydata(self):
        self.gotText(self.bodydata)
        self.bodydata = ""

    def do_waitforendscript(self, byte):
        if byte == "<":
            return "waitscriptendtag"
        self.bodydata += byte

    def begin_waitscriptendtag(self, byte):
        self.temptagdata = ""
        self.tagName = ""
        self.endtag = 0

    def do_waitscriptendtag(self, byte):
        # 1 enforce / as first byte read
        # 2 enforce following bytes to be subset of "script" until
        #   tagName == "script"
        #   2a when that happens, gotText(self.bodydata) and
        #      gotTagEnd(self.tagName)
        # 3 spaces can happen anywhere, they're ignored
        #   e.g. < / script >
        # 4 anything else causes all data I've read to be moved to the
        #   bodydata, and switch back to waitforendscript state

        # If it turns out this _isn't_ a </script>, we need to
        # remember all the data we've been through so we can append it
        # to bodydata
        self.temptagdata += byte

        # 1
        if byte == "/":
            self.endtag = True
        elif not self.endtag:
            self.bodydata += "<" + self.temptagdata
            return "waitforendscript"
        # 2
        elif byte.isalnum() or byte in IDENTCHARS:
            self.tagName += byte
            if not "script".startswith(self.tagName):
                self.bodydata += "<" + self.temptagdata
                return "waitforendscript"
            elif self.tagName == "script":
                self.gotText(self.bodydata)
                self.gotTagEnd(self.tagName)
                return "waitforgt"
        # 3
        elif byte.isspace():
            return "waitscriptendtag"
        # 4
        else:
            self.bodydata += "<" + self.temptagdata
            return "waitforendscript"

    def begin_entityref(self, byte):
        self.erefbuf = ""
        self.erefextra = ""  # extra bit for lenient mode

    def do_entityref(self, byte):
        if byte.isspace() or byte == "<":
            if self.lenient:
                # '&foo' probably was '&amp;foo'
                if self.erefbuf and self.erefbuf != "amp":
                    self.erefextra = self.erefbuf

                self.erefbuf = "amp"

                if byte == "<":
                    return "tagstart"
                else:
                    self.erefextra += byte
                    return "spacebodydata"

            self._raise_parse_error("Bad entity reference")
        elif byte != ";":
            self.erefbuf += byte
        else:
            return "bodydata"

    def end_entityref(self):
        self.gotEntityReference(self.erefbuf)

    # hacky support for space after & in entityref in lenient
    # state should only happen in that case
    def begin_spacebodydata(self, byte):
        self.bodydata = self.erefextra
        self.erefextra = None

    do_spacebodydata = do_bodydata
    end_spacebodydata = end_bodydata

    # Sorta SAX-ish API

    def gotTagStart(self, name, attributes):
        """Encountered an opening tag.

        Default behaviour is to print."""
        print("begin", name, attributes)

    def gotText(self, data):
        """Encountered text

        Default behaviour is to print."""
        print("text:", repr(data))

    def gotEntityReference(self, entityRef):
        """Encountered mnemonic entity reference

        Default behaviour is to print."""
        print("entityRef: &%s;" % entityRef)

    def gotComment(self, comment):
        """Encountered comment.

        Default behaviour is to ignore."""
        pass

    def gotCData(self, cdata):
        """Encountered CDATA

        Default behaviour is to call the gotText method"""
        self.gotText(cdata)

    def gotDoctype(self, doctype):
        """Encountered DOCTYPE

        This is really grotty: it basically just gives you everything between
        '<!DOCTYPE' and '>' as an argument.
        """
        print("!DOCTYPE", repr(doctype))

    def gotTagEnd(self, name):
        """Encountered closing tag

        Default behaviour is to print."""
        print("end", name)



from pprint import pprint
from riko.bado import coroutine
from riko.collections import SyncPipe, AsyncPipe

p120_conf = {"type": "text"}
p120_inputs = {"format": "%B %d, %Y"}
p112_conf = {"type": "date", "default": "5/4/82", "prompt": "enter a date"}
p151_conf = {"format": {"terminal": "format", "path": "format"}}
p100_conf = {
    "attrs": {"value": {"terminal": "value", "path": "dateformat"}, "key": "date"}
}

p120_kwargs = {"conf": p120_conf, "inputs": p120_inputs, "assign": "format"}


def pipe(test=False):
    s1 = SyncPipe("input", test=test, **p120_kwargs).output
    s2 = (
        SyncPipe("input", conf=p112_conf, test=test)
        .dateformat(conf=p151_conf, format=s1)
        .output
    )

    stream = SyncPipe("itembuilder", conf=p100_conf, value=s2, test=test).list

    for i in stream:
        pprint(i)

    return stream


@coroutine
def async_pipe(reactor, test=False):
    s1 = yield AsyncPipe("input", test=test, **p120_kwargs).output
    s2 = yield (
        AsyncPipe("input", conf=p112_conf, test=test)
        .dateformat(conf=p151_conf, format=s1)
        .output
    )

    output_kwargs = {"conf": p100_conf, "value": s2, "test": test}
    output = yield (AsyncPipe("itembuilder", **output_kwargs).list)

    for i in output:
        pprint(i)
